import copy
import numpy as np
import re

from PySide6.QtWidgets import (
    QDialog,
    QWidget,
    QDialogButtonBox,
    QGridLayout,
    QLabel,
    QMessageBox,
    QPushButton,
    QCheckBox,
)
from PySide6.QtCore import Qt
from PySide6.QtGui import QFont, QKeyEvent

from lmfit import models

from nomad_camels.gui.plot_definer import Ui_Plot_Definer
from nomad_camels.gui.plot_definer_2d import Ui_Plot_Definer_2D
from nomad_camels.gui.fit_definer import Ui_Fit_Definer

from nomad_camels.ui_widgets.add_remove_table import AddRemoveTable
from nomad_camels.utility import variables_handling
from nomad_camels.utility.fit_variable_renaming import replace_name


plot_types = ["X-Y plot", "Value-List", "2D plot"]


class Plot_Info:
    """ """

    def __init__(
        self,
        plt_type="X-Y plot",
        x_axis="",
        y_axes=None,
        title="",
        xlabel="",
        ylabel="",
        ylabel2="",
        do_plot=True,
        zlabel="",
        same_fit=False,
        fits=None,
        all_fit=None,
        z_axis="",
        logX=False,
        logY=False,
        logY2=False,
    ):
        self.plt_type = plt_type
        self.x_axis = x_axis
        self.y_axes = y_axes or {"formula": [], "axis": []}
        self.z_axis = z_axis
        self.title = title
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.ylabel2 = ylabel2
        self.logX = logX
        self.logY = logY
        self.logY2 = logY2
        self.zlabel = zlabel
        self.do_plot = do_plot
        self.same_fit = same_fit
        self.fits = fits or []
        self.plot_all_available = True
        self.all_fit = all_fit or Fit_Info()
        self.name = ""
        self.maxlen = np.inf
        self.update_name()

    def update_name(self):
        """ """
        if self.title:
            self.name = self.title
        elif self.plt_type == "X-Y plot":
            if self.xlabel and self.ylabel:
                self.name = f"{self.ylabel} vs. {self.xlabel}"
            elif self.x_axis and self.y_axes["formula"]:
                self.name = f'{self.y_axes["formula"][0]} vs. {self.x_axis}'
            elif self.y_axes["formula"]:
                self.name = self.y_axes["formula"][0]
            else:
                self.name = self.x_axis
        elif self.plt_type == "Value-List":
            self.name = "Current Values"
        elif self.plt_type == "2D plot":
            if self.zlabel and self.xlabel and self.ylabel:
                self.name = f"{self.zlabel} vs. ({self.xlabel}; {self.ylabel})"
            elif self.zlabel:
                self.name = f"{self.zlabel} 2D"
            if self.z_axis and self.x_axis and self.y_axes["formula"]:
                self.name = (
                    f'{self.z_axis} vs. ({self.x_axis}; {self.y_axes["formula"][0]})'
                )
            elif self.z_axis:
                self.name = f"{self.z_axis} 2D"

    def get_fit_vars(self, stream=""):
        """

        Parameters
        ----------
        stream :
             (Default value = '')

        Returns
        -------

        """
        variables = {}
        if self.same_fit:
            if self.all_fit.do_fit:
                self.all_fit.x = self.x_axis
                for y in self.y_axes["formula"]:
                    fit = copy.deepcopy(self.all_fit)
                    fit.y = y
                    variables.update(fit.get_variables(stream))
        else:
            for fit in self.fits:
                if fit.do_fit:
                    variables.update(fit.get_variables(stream))
        return variables


class Fit_Info:
    """ """

    def __init__(
        self,
        do_fit=False,
        predef_func="",
        custom_func="",
        use_custom_func=False,
        guess_params=True,
        initial_params=None,
        y="",
        x="",
        additional_data=None,
        display_values=False,
    ):
        self.do_fit = do_fit
        self.predef_func = predef_func
        self.custom_func = custom_func
        self.use_custom_func = use_custom_func
        self.guess_params = guess_params
        self.name = ""
        self.display_values = display_values
        self.initial_params = initial_params or {
            "name": [],
            "initial value": [],
            "lower bound": [],
            "upper bound": [],
        }
        self.additional_data = additional_data or []
        self.y = y or ""
        self.x = x or ""

    def get_name(self, stream=""):
        """

        Parameters
        ----------
        stream :
             (Default value = '')

        Returns
        -------

        """
        if self.use_custom_func:
            label = "custom"
        else:
            label = self.predef_func
        self.name = f"{label}_{self.y}_v_{self.x}_{stream}".replace(" ", "_")
        return self.name

    def get_variables(self, stream=""):
        """

        Parameters
        ----------
        stream :
             (Default value = '')

        Returns
        -------

        """
        variables = {}
        name = self.get_name(stream)
        for var in self.initial_params["name"]:
            var_name = f"{name}_{var}"
            var_name = replace_name(var_name)
            variables[var_name] = 1
        return variables


class Plot_Definer(QDialog):
    """ """

    def __init__(self, parent=None, plot_data=None):
        self.plot_data = plot_data or []
        super().__init__(parent)
        self.setWindowTitle("Define plot - NOMAD CAMELS")
        self.setWindowFlags(self.windowFlags() | Qt.WindowMaximizeButtonHint)
        cols = ["plot-type", "name"]
        comboBoxes = {"plot-type": plot_types}
        tableData = {"plot-type": [], "name": []}
        for plt in self.plot_data:
            tableData["plot-type"].append(plt.plt_type)
            tableData["name"].append(plt.name)
        self.plot_table = AddRemoveTable(
            headerLabels=cols,
            title="Plots",
            editables=[],
            comboBoxes=comboBoxes,
            tableData=tableData,
            askdelete=True,
        )
        self.plot_table.table.clicked.connect(self.change_plot_def)
        self.plot_table.added.connect(self.plot_added)
        self.plot_table.removed.connect(self.plot_removed)
        self.plot_def = QLabel("Select a plot!")
        self.dialog_buttons = QDialogButtonBox()
        self.dialog_buttons.setOrientation(Qt.Horizontal)
        self.dialog_buttons.setStandardButtons(
            QDialogButtonBox.Cancel | QDialogButtonBox.Ok
        )
        self.dialog_buttons.accepted.connect(self.accept)
        self.dialog_buttons.rejected.connect(self.reject)
        self.setWindowFlags(self.windowFlags() | Qt.WindowMaximizeButtonHint)

        layout = QGridLayout()
        layout.addWidget(self.plot_table, 0, 0)
        layout.addWidget(self.plot_def, 0, 1)
        layout.addWidget(self.dialog_buttons, 1, 0, 1, 2)
        self.setLayout(layout)

    def accept(self) -> None:
        """ """
        if not isinstance(self.plot_def, QLabel):
            self.plot_def.get_data()
        super().accept()

    def reject(self):
        """ """
        discard_dialog = QMessageBox.question(
            self,
            "Discard Changes?",
            f"All changes to the defined plots / fits will be lost!",
            QMessageBox.Yes | QMessageBox.No,
        )
        if discard_dialog != QMessageBox.Yes:
            return
        super().reject()

    def change_plot_def(self, a0):
        """

        Parameters
        ----------
        a0 :


        Returns
        -------

        """
        if not isinstance(self.plot_def, QLabel):
            self.plot_def.get_data()
        plot_dat = self.plot_data[a0.row()]
        ind = self.plot_table.table_model.index(a0.row(), 0)
        plot_dat.plt_type = self.plot_table.table.indexWidget(ind).currentText()
        tableData = {"plot-type": [], "name": []}
        for plt in self.plot_data:
            tableData["plot-type"].append(plt.plt_type)
            tableData["name"].append(plt.name)
        self.plot_table.change_table_data(tableData)
        if plot_dat.plt_type == "X-Y plot":
            plot_def = Single_Plot_Definer_XY(plot_dat, self)
        elif plot_dat.plt_type == "Value-List":
            plot_def = Single_Plot_Definer_List(plot_dat, self)
        elif plot_dat.plt_type == "2D plot":
            plot_def = Single_Plot_Definer_2D(plot_dat, self)
        else:
            plot_def = QLabel("Not implemented yet!")
        self.layout().replaceWidget(self.plot_def, plot_def)
        self.plot_def.deleteLater()
        self.plot_def = plot_def
        flag = self.plot_table.table.selectionModel().SelectionFlag.Select
        self.plot_table.table.selectionModel().select(a0, flag)

    def plot_added(self, n):
        """

        Parameters
        ----------
        n :


        Returns
        -------

        """
        if n >= len(self.plot_data):
            self.plot_data.append(Plot_Info())

    def plot_removed(self, n):
        """

        Parameters
        ----------
        n :


        Returns
        -------

        """
        self.plot_data.pop(n)


class Single_Plot_Definer(QWidget):
    """ """

    def __init__(self, plot_data: Plot_Info, parent=None):
        super().__init__(parent)
        self.plot_data = plot_data

    def get_data(self):
        """ """
        return self.plot_data

    def add_y(self, n):
        """

        Parameters
        ----------
        n :


        Returns
        -------

        """
        if n >= len(self.plot_data.fits):
            self.plot_data.fits.append(Fit_Info())

    def remove_y(self, n):
        """

        Parameters
        ----------
        n :


        Returns
        -------

        """
        self.plot_data.fits.pop(n)


class Single_Plot_Definer_List(Single_Plot_Definer):
    """ """

    def __init__(self, plot_data: Plot_Info, parent=None):
        super().__init__(plot_data, parent)
        self.plot_data = plot_data
        self.checkBox_plot_all = QCheckBox("plot all available channels")
        self.checkBox_plot_all.setChecked(plot_data.plot_all_available)
        self.table = AddRemoveTable(
            title="Values",
            headerLabels=[],
            tableData=plot_data.y_axes["formula"],
            checkstrings=[0],
        )
        self.table.added.connect(self.add_y)
        self.table.removed.connect(self.remove_y)
        self.layout = QGridLayout()
        self.layout.addWidget(self.checkBox_plot_all)
        self.layout.addWidget(self.table)
        self.setLayout(self.layout)
        self.plot_data.update_name()

    def get_data(self):
        """ """
        self.plot_data.plot_all_available = self.checkBox_plot_all.isChecked()
        self.plot_data.y_axes["formula"] = self.table.update_table_data()
        self.plot_data.y_axes["axis"] = [1] * len(self.plot_data.y_axes["formula"])
        self.plot_data.update_name()
        return super().get_data()


class Single_Plot_Definer_2D(Ui_Plot_Definer_2D, Single_Plot_Definer):
    """ """

    def __init__(self, plot_data: Plot_Info, parent=None):
        super().__init__(plot_data, parent)
        self.setupUi(self)
        self.lineEdit_x_axis.setText(self.plot_data.x_axis)
        if not self.plot_data.y_axes["formula"]:
            self.plot_data.y_axes["formula"].append("")
            self.plot_data.y_axes["axis"].append(1)
        self.lineEdit_y_axis.setText(self.plot_data.y_axes["formula"][0])
        self.lineEdit_z_axis.setText(self.plot_data.z_axis)
        self.lineEdit_xlabel.setText(self.plot_data.xlabel)
        self.lineEdit_ylabel.setText(self.plot_data.ylabel)
        self.lineEdit_zlabel.setText(self.plot_data.zlabel)
        self.lineEdit_title.setText(self.plot_data.title)
        self.lineEdit_n_data_points.setText(str(self.plot_data.maxlen))
        self.plot_data.update_name()

    def get_data(self):
        """ """
        self.plot_data.xlabel = self.lineEdit_xlabel.text()
        if re.search(r"[\\'\"`]", self.plot_data.xlabel):
            raise ValueError(
                "x-label contains special characters.\nYou cannot use ' \" or `."
            )
        self.plot_data.ylabel = self.lineEdit_ylabel.text()
        if re.search(r"[\\'\"`]", self.plot_data.ylabel):
            raise ValueError(
                "y-label contains special characters.\nYou cannot use ' \" or `."
            )
        self.plot_data.zlabel = self.lineEdit_zlabel.text()
        if re.search(r"[\\'\"`]", self.plot_data.zlabel):
            raise ValueError(
                "z-label contains special characters.\nYou cannot use ' \" or `."
            )
        self.plot_data.x_axis = self.lineEdit_x_axis.text()
        self.plot_data.y_axes["formula"][0] = self.lineEdit_y_axis.text()
        self.plot_data.z_axis = self.lineEdit_z_axis.text()
        self.plot_data.title = self.lineEdit_title.text()
        self.plot_data.maxlen = self.lineEdit_n_data_points.text() or "inf"
        self.plot_data.update_name()
        return super().get_data()


class Single_Plot_Definer_XY(Ui_Plot_Definer, Single_Plot_Definer):
    """ """

    def __init__(self, plot_data: Plot_Info, parent=None):
        super().__init__(plot_data, parent)
        self.fit_definer = None
        self.setupUi(self)

        cols = ["formula", "axis"]
        comboBoxes = {"axis": ["left", "right"]}
        self.y_table = AddRemoveTable(
            headerLabels=cols,
            title="y-axes",
            comboBoxes=comboBoxes,
            tableData=plot_data.y_axes,
            checkstrings=[0],
        )
        self.y_table.added.connect(self.add_y)
        self.y_table.removed.connect(self.remove_y)
        self.y_table.table.clicked.connect(self.fit_change)
        self.checkBox_same_fit.clicked.connect(self.fit_change)
        self.layout().replaceWidget(self.y_axes, self.y_table)

        # self.checkBox_plot.clicked.connect(self.plot_change)
        self.load_data()
        # self.plot_change()
        self.fit_change()

    # def plot_change(self):
    #     """ """
    #     self.plotting_group.setEnabled(self.checkBox_plot.isChecked())

    def load_data(self):
        """ """
        self.lineEdit_x_axis.setText(self.plot_data.x_axis)
        self.lineEdit_title.setText(self.plot_data.title)
        self.lineEdit_xlabel.setText(self.plot_data.xlabel)
        self.lineEdit_ylabel.setText(self.plot_data.ylabel)
        self.lineEdit_ylabel2.setText(self.plot_data.ylabel2)
        # self.checkBox_plot.setChecked(self.plot_data.do_plot)
        self.lineEdit_nPoints.setText(str(self.plot_data.maxlen))
        self.checkBox_same_fit.setChecked(self.plot_data.same_fit)
        self.checkBox_xlog.setChecked(self.plot_data.logX)
        self.checkBox_ylog.setChecked(self.plot_data.logY)
        self.checkBox_ylog2.setChecked(self.plot_data.logY2)
        self.plot_data.update_name()

    def get_data(self):
        """ """
        self.plot_data.y_axes = self.y_table.update_table_data()
        self.plot_data.x_axis = self.lineEdit_x_axis.text()
        self.plot_data.title = self.lineEdit_title.text()
        if re.search(r"[\\'\"`]", self.plot_data.title):
            raise ValueError(
                "Title contains special characters.\nYou cannot use ' \" or `."
            )
        self.plot_data.xlabel = self.lineEdit_xlabel.text()
        if re.search(r"[\\'\"`]", self.plot_data.xlabel):
            raise ValueError(
                "x-label contains special characters.\nYou cannot use ' \" or `."
            )
        self.plot_data.ylabel = self.lineEdit_ylabel.text()
        if re.search(r"[\\'\"`]", self.plot_data.ylabel):
            raise ValueError(
                "y-label contains special characters.\nYou cannot use ' \" or `."
            )
        self.plot_data.ylabel2 = self.lineEdit_ylabel2.text()
        if re.search(r"[\\'\"`]", self.plot_data.zlabel):
            raise ValueError(
                "z-label contains special characters.\nYou cannot use ' \" or `."
            )
        # self.plot_data.do_plot = self.checkBox_plot.isChecked()
        try:
            self.plot_data.maxlen = int(self.lineEdit_nPoints.text())
        except ValueError:
            self.plot_data.maxlen = np.inf
        self.plot_data.same_fit = self.checkBox_same_fit.isChecked()
        self.plot_data.logX = self.checkBox_xlog.isChecked()
        self.plot_data.logY = self.checkBox_ylog.isChecked()
        self.plot_data.logY2 = self.checkBox_ylog2.isChecked()
        if not isinstance(self.fit_definer, QLabel):
            self.fit_definer.get_data()
        self.plot_data.update_name()
        if self.plot_data.all_fit:
            self.plot_data.all_fit.x = self.plot_data.x_axis
        for i, fit in enumerate(self.plot_data.fits):
            fit.y = self.plot_data.y_axes["formula"][i]
            fit.x = self.plot_data.x_axis
        return super().get_data()

    def fit_change(self):
        """ """
        if not isinstance(self.fit_definer, QLabel):
            self.fit_definer.get_data()
        if self.checkBox_same_fit.isChecked():
            fit_dat = self.plot_data.all_fit
            fit_to = "all y-axes"
        else:
            ind = self.y_table.table.selectedIndexes()
            if ind:
                n = ind[0].row()
                fit_to = self.y_table.table_model.item(n, 0).text()
                fit_dat = self.plot_data.fits[n]
            else:
                fit_dat = None
                fit_to = None
        if fit_dat is None:
            fit_definer = QLabel("No y-axis selected")
        else:
            fit_definer = Fit_Definer(fit_dat, self, fit_to)
        self.layout().replaceWidget(self.fit_definer, fit_definer)
        self.fit_definer.deleteLater()
        self.fit_definer = fit_definer


class Fit_Definer(Ui_Fit_Definer, QWidget):
    """ """

    def __init__(self, fit_info: Fit_Info, parent=None, fit_to=""):
        super().__init__(parent)
        self.setupUi(self)
        self.fit_info = fit_info
        self.label.setText(f"Fit to: {fit_to}")
        self.comboBox_predef_func.addItems(sorted(models_names.keys()))
        cols = ["name", "initial value", "lower bound", "upper bound"]
        self.start_params = AddRemoveTable(
            headerLabels=cols,
            title="Fit Parameters",
            editables=[1, 2, 3],
            tableData=fit_info.initial_params,
        )
        self.add_data = AddRemoveTable(
            headerLabels=[], title="Additional Data", tableData=fit_info.additional_data
        )
        self.start_params.addButton.setHidden(True)
        self.start_params.removeButton.setHidden(True)
        self.load_data()

        self.checkBox_fit.clicked.connect(self.change_func)
        self.radioButton_custom_func.clicked.connect(self.change_func)
        self.radioButton_predef_func.clicked.connect(self.change_func)
        self.checkBox_guess.clicked.connect(self.change_func)
        self.comboBox_predef_func.currentTextChanged.connect(self.change_func)
        self.lineEdit_custom_func.textChanged.connect(self.change_func)
        self.change_func()
        self.layout().addWidget(self.start_params, 10, 0, 1, 2)
        self.layout().addWidget(self.add_data, 0, 3, 11, 2)

    def change_func(self):
        """ """
        fit = self.checkBox_fit.isChecked()
        custom = self.radioButton_custom_func.isChecked()
        guess = self.checkBox_guess.isChecked()
        self.radioButton_custom_func.setEnabled(fit)
        self.radioButton_predef_func.setEnabled(fit)
        self.checkBox_guess.setEnabled(fit)
        self.comboBox_predef_func.setEnabled(fit and not custom)
        self.lineEdit_custom_func.setEnabled(fit and custom)
        # self.start_params.setEnabled(fit and not guess)
        self.start_params.enableds = [0, 2, 3]
        self.start_params.editables = [2, 3]
        self.start_params.addButton.setHidden(not custom)
        self.start_params.removeButton.setHidden(not custom)
        if custom:
            self.start_params.editables += [0]
        if not guess:
            self.start_params.editables += [1]
            self.start_params.enableds += [1]
        par_vals = self.start_params.update_table_data()
        self.start_params.change_table_data(par_vals)

        func = self.comboBox_predef_func.currentText()
        if func != self.fit_info.predef_func:
            if custom:
                try:
                    mod = models.ExpressionModel(self.lineEdit_custom_func.text())
                except (ValueError, SyntaxError):
                    self.lineEdit_custom_func.setStyleSheet(
                        f'background-color: rgb{variables_handling.get_color("red", True)}'
                    )
                    return
                self.lineEdit_custom_func.setStyleSheet(
                    f'background-color: rgb{variables_handling.get_color("green", True)}'
                )
            else:
                mod = models_names[func]()
            params = mod.param_names
            par_vals = {
                "name": [],
                "initial value": [],
                "lower bound": [],
                "upper bound": [],
            }
            for param in params:
                par_vals["name"].append(param)
                par_vals["initial value"].append(1)
                par_vals["lower bound"].append("")
                par_vals["upper bound"].append("")
            self.start_params.change_table_data(par_vals)
            self.fit_info.predef_func = func
        if fit and not custom and func == "Linear" and par_vals["name"] == []:
            mod = models_names[func]()
            params = mod.param_names
            par_vals = {
                "name": [],
                "initial value": [],
                "lower bound": [],
                "upper bound": [],
            }
            for param in params:
                par_vals["name"].append(param)
                par_vals["initial value"].append(1)
                par_vals["lower bound"].append("")
                par_vals["upper bound"].append("")
            self.start_params.change_table_data(par_vals)


    def load_data(self):
        """ """
        self.checkBox_fit.setChecked(self.fit_info.do_fit)
        self.comboBox_predef_func.setCurrentText(self.fit_info.predef_func)
        self.lineEdit_custom_func.setText(self.fit_info.custom_func)
        self.radioButton_custom_func.setChecked(self.fit_info.use_custom_func)
        self.checkBox_guess.setChecked(self.fit_info.guess_params)
        self.checkBox_display_values.setChecked(self.fit_info.display_values)

    def get_data(self):
        """ """
        self.fit_info.initial_params = self.start_params.update_table_data()
        rems = []
        for i, param in enumerate(self.fit_info.initial_params["name"]):
            if not param:
                rems.append(i)
        for i in rems[::-1]:
            for key in self.fit_info.initial_params:
                self.fit_info.initial_params[key].pop(i)
        self.fit_info.additional_data = self.add_data.update_table_data()
        self.fit_info.do_fit = self.checkBox_fit.isChecked()
        self.fit_info.predef_func = self.comboBox_predef_func.currentText()
        self.fit_info.custom_func = self.lineEdit_custom_func.text()
        self.fit_info.use_custom_func = self.radioButton_custom_func.isChecked()
        self.fit_info.guess_params = self.checkBox_guess.isChecked()
        self.fit_info.display_values = self.checkBox_display_values.isChecked()


def add_fit(tableData, fit):
    """

    Parameters
    ----------
    tableData :

    fit :


    Returns
    -------

    """
    custom = fit.use_custom_func
    if custom:
        tableData["fit"].append(fit.custom_func)
    else:
        tableData["fit"].append(fit.predef_func)


def make_table_data(plot_data):
    """

    Parameters
    ----------
    plot_data :


    Returns
    -------

    """
    tableData = {"plot-type": [], "name": [], "fit": []}
    for plt in plot_data:
        tableData["plot-type"].append(plt.plt_type)
        tableData["name"].append(plt.name)
        if plt.all_fit and plt.all_fit.do_fit and plt.same_fit:
            add_fit(tableData, plt.all_fit)
        elif plt.fits:
            added = False
            for fit in plt.fits:
                if not fit.do_fit:
                    continue
                add_fit(tableData, fit)
                added = True
                break
            if not added:
                tableData["fit"].append("None")
        else:
            tableData["fit"].append("None")
    return tableData


class Plot_Button_Overview(QWidget):
    """ """

    def __init__(self, parent, plot_data=None):
        self.plot_data = plot_data
        super().__init__(parent)
        tableData = make_table_data(plot_data)
        cols = ["plot-type", "name", "fit"]
        self.plot_table = AddRemoveTable(
            headerLabels=cols,
            title="plot-overview",
            editables=[],
            tableData=tableData,
            askdelete=True,
            horizontal=False,
        )
        self.plot_table.addButton.setHidden(True)
        self.plot_table.removeButton.setHidden(True)
        self.plot_button = QPushButton("Configure: Plots/Fits")
        font = QFont()
        font.setBold(True)
        self.plot_button.setStyleSheet(
            """
            QPushButton {
                background-color: #4CAF50; /* green */
                border: none; /* Remove border for a cleaner look */
                color: white; /* Text color */
                padding: 3px; /* Adjust padding as needed */
                padding-bottom: 3px;
                border-radius: 6px; /* Rounded corners with a radius of 10px */
                font-size: 13px;
                font-weight: bold;
                text-align: center;
                margin: 2px 2px;
            }
            QPushButton:hover {
                background-color: #45a049; /* Slightly lighter green on hover */
            }
            QPushButton:pressed {
                background-color: #2f6331; /* Slightly darker green when pressed */
            }
            """
        )
        self.plot_button.setFont(font)
        self.plot_button.clicked.connect(self.define_plots)

        layout = QGridLayout()
        layout.addWidget(self.plot_button, 0, 0)
        layout.addWidget(self.plot_table, 1, 0)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    def define_plots(self):
        """ """
        plot_definer = Plot_Definer(self, self.plot_data)
        if plot_definer.exec():
            self.plot_data = plot_definer.plot_data
            tableData = make_table_data(self.plot_data)
            self.plot_table.change_table_data(tableData)

    def keyPressEvent(self, a0: QKeyEvent) -> None:
        """Overwrites the keyPressEvent of the QDialog so that it does
        not close when pressing Enter/Return.

        Parameters
        ----------
        a0: QKeyEvent :


        Returns
        -------

        """
        if a0.key() == Qt.Key_Enter or a0.key() == Qt.Key_Return:
            return
        super().keyPressEvent(a0)


models_names = dict(models.lmfit_models)
models_names.pop("Expression")



from PySide6.QtWidgets import QCheckBox

from nomad_camels.main_classes.loop_step import Loop_Step, Loop_Step_Config

from nomad_camels.ui_widgets.channels_check_table import Channels_Check_Table
from nomad_camels.utility import variables_handling


class Set_Channels(Loop_Step):
    """
    Simple step to set several channels to a given value. Uses `bps.abs_set`.

    Attributes
    ----------
    channels_values : dict{'Channels': list[str], 'Values': list[str]}
        This dictionary provides a list of the names of the channels that are to
        be set and a list of the values they should get. The values are strings,
        since they will be evaluated at runtime, thus providing the ability to
        set some variable at runtime.
    wait_for_set : bool, default True
        Whether to wait after setting for the set channels to have the finished
        status.
    """

    def __init__(self, name="", parent_step=None, step_info=None, **kwargs):
        super().__init__(name, parent_step, step_info, **kwargs)
        self.step_type = "Set Channels"
        if step_info is None:
            step_info = {}
        self.channels_values = (
            step_info["channels_values"]
            if "channels_values" in step_info
            else {"Channels": [], "Values": []}
        )
        self.wait_for_set = (
            step_info["wait_for_set"] if "wait_for_set" in step_info else True
        )
        self.update_used_devices()

    def update_used_devices(self):
        """All devices with a channel that is to be set are added."""
        self.used_devices = []
        for channel in self.channels_values["Channels"]:
            if channel in variables_handling.channels:
                device = variables_handling.channels[channel].device
                if device not in self.used_devices:
                    self.used_devices.append(device)

    def get_protocol_string(self, n_tabs=1):
        """If `wait_for_set` is True, then after setting, bps.wait for
        the set group is called. In any case, all the channels are set to their
        specified value"""
        tabs = "\t" * n_tabs
        protocol_string = super().get_protocol_string(n_tabs)
        for i, channel in enumerate(self.channels_values["Channels"]):
            if channel not in variables_handling.channels:
                raise Exception(
                    f"Trying to set channel {channel} in {self.full_name}, but it does not exist!"
                )
            dev, chan = variables_handling.channels[channel].name.split(".")
            val = self.channels_values["Values"][i]
            protocol_string += f'{tabs}yield from bps.abs_set(devs["{dev}"].{chan}, eva.eval("{val}"), group="A")\n'
        if self.wait_for_set:
            protocol_string += f'{tabs}yield from bps.wait("A")\n'
        return protocol_string

    def get_protocol_short_string(self, n_tabs=0):
        """Displays the channels and their values."""
        short_string = super().get_protocol_short_string(n_tabs)
        val_dic = {self.channels_values["Channels"][i]: self.channels_values["Values"][i].replace('"', '\'') for i in range(len(self.channels_values["Channels"]))}
        short_string = f"{short_string[:-1]} - {val_dic}\n"
        return short_string


class Set_Channels_Config(Loop_Step_Config):
    """The configuration consists of the checkbox for waiting and a
    simple AddRemoveTable that works with the channels.

    Parameters
    ----------

    Returns
    -------

    """

    def __init__(self, loop_step: Set_Channels, parent=None):
        super().__init__(parent, loop_step)
        box = []
        for channel in variables_handling.channels:
            if variables_handling.channels[channel].output:
                box.append(channel)
        info_dict = {
            "channel": self.loop_step.channels_values["Channels"],
            "value": self.loop_step.channels_values["Values"],
        }
        # self.sub_widget = AddRemoveTable(headerLabels=['Channels', 'Values'],
        #                                  comboBoxes={'Channels': box},
        #                                  tableData=loop_step.channels_values,
        #                                  checkstrings=1)
        self.sub_widget = Channels_Check_Table(
            self, ["set", "channel", "value"], True, info_dict, [2]
        )
        self.checkBox_wait_for_set = QCheckBox("Wait for set")
        self.checkBox_wait_for_set.setChecked(True)
        self.checkBox_wait_for_set.stateChanged.connect(self.check_change)
        self.layout().addWidget(self.checkBox_wait_for_set, 1, 0, 1, 5)
        self.layout().addWidget(self.sub_widget, 2, 0, 1, 5)

    def check_change(self):
        """ """
        self.loop_step.wait_for_set = self.checkBox_wait_for_set.isChecked()

    def update_step_config(self):
        """ """
        super().update_step_config()
        info = self.sub_widget.get_info()
        self.loop_step.channels_values = {
            "Channels": info["channel"],
            "Values": info["value"],
        }
        # self.sub_widget.update_table_data()
        # self.loop_step.channels_values = self.sub_widget.tableData



"""This module helps to place the plots on the screen without too much
overlap between the plots. The positions are saved into the global variables,
since this module may be imported by several other modules.

Attributes
----------
    current_screen : int
        the number of the currently used screen to place plots
    current_pos : list[int, int]
        the x and y coordinates of the current placement
    max_height_in_row : int
        the maximum height a plot window has within the currently populated row
    iteration : int
        if the screens were full once, it is increased and the next plots are
        slightly shifted
    horizontal_margin : int
        the horizontal distance between the plot windows
    vertical_margin : int
        the vertical distance between the plot windows
"""

import sys

from PySide6.QtWidgets import QWidget, QApplication
from PySide6.QtCore import QCoreApplication

app = QCoreApplication.instance()
if app is None:
    app = QApplication(sys.argv)
screens = app.screens()

current_screen = 0
current_pos = [0, 0]
max_height_in_row = 0
iteration = 0

horizontal_margin = 5
vertical_margin = 5


def reset_variables():
    global current_screen, current_pos, max_height_in_row, iteration, screens
    screens = app.screens()
    current_screen = 0
    current_pos = [0, 0]
    max_height_in_row = 0
    iteration = 0
    screen_geometry = screens[current_screen].availableGeometry()


def place_widget(widget: QWidget):
    """
    This function places the given `widget` on the next free spot on the screen.
    First all widgets are placed next to each other in a row, until one would be
    outside the screen, then the next row is started. If it would "flow out" the
    bottom of the screen, a second screen (if available) is used. If all
    available screens are full, it restarts on the first screen, slightly
    shifting the widgets in comparison to the first ones.

    Parameters
    ----------
    widget : QWidget
        the widget that should be placed
    """
    global max_height_in_row, current_screen, current_pos, iteration, screens
    widget.show()
    c = current_pos
    # print(c, current_screen)
    s = widget.size()
    try:
        screen_geometry = screens[current_screen].availableGeometry()
    except:
        reset_variables()
        screen_geometry = screens[current_screen].availableGeometry()
    if current_pos[0] + s.width() > screen_geometry.width():
        current_pos[1] += max_height_in_row + horizontal_margin
        current_pos[0] = 0
        max_height_in_row = 0
        if current_pos[1] + s.height() > screen_geometry.height():
            current_pos[1] = 0
            current_screen += 1
            if current_screen >= len(screens):
                current_screen = 0
                iteration += 1
            current_pos[1] += 5 * iteration * vertical_margin
    widget.move(
        current_pos[0] + screen_geometry.x(), current_pos[1] + screen_geometry.y()
    )
    # print(current_pos[0] + screen_geometry.x(), current_pos[1] + screen_geometry.y())
    current_pos[0] += s.width() + vertical_margin
    max_height_in_row = (
        s.height() if s.height() > max_height_in_row else max_height_in_row
    )



from traceback import format_tb, print_tb, format_exception
from PySide6.QtWidgets import QMessageBox, QTextEdit, QPushButton, QVBoxLayout
from PySide6.QtCore import QUrl

from PySide6.QtMultimedia import QSoundEffect

import logging

from bluesky.utils import RunEngineInterrupted

from nomad_camels.utility import variables_handling
from importlib import resources
from nomad_camels import graphics

# imported, so that it is run once, before the logging in
# `exception_hook` is connected
from nomad_camels.utility import logging_settings


class ErrorMessage(QMessageBox):
    """A popUp-box describing an Error.

    Parameters
    ----------
    msg : str
        the error message
    info_text : str
        A longer text, explaining the error (usually traceback)
    parent : QWidget
        The parent widget of this Messagebox
    """

    def __init__(self, exc_info, parent=None):
        tb = format_tb(exc_info[2])
        formatted_exc = "".join(format_exception(exc_info[0], exc_info[1], exc_info[2]))
        print_tb(exc_info[2])
        msg = str(exc_info[0].__name__)
        info_text = str(exc_info[1])
        logging.exception(formatted_exc)
        print(msg)
        print(info_text)
        super().__init__(parent)
        self.setWindowTitle("ERROR")
        self.setIcon(QMessageBox.Warning)
        self.setStandardButtons(QMessageBox.Ok)
        self.setText(msg)
        if info_text:
            self.setInformativeText(info_text)

        self.more_info_button = QPushButton("more information")
        self.more_info_button.clicked.connect(self.show_more_info)

        self.text_edit = QTextEdit()
        self.text_edit.setReadOnly(True)
        self.text_edit.setVisible(False)
        self.text_edit.setPlainText("".join(tb))

        self.layout().addWidget(self.more_info_button, 3, 0, 1, 5)
        self.layout().addWidget(self.text_edit, 2, 0, 1, 5)
        self.layout().addWidget(self.button(QMessageBox.Ok), 4, 0, 1, 5)
        self.showing_more_info = False
        self.adjustSize()

    def show_more_info(self):
        if self.showing_more_info:
            self.text_edit.setVisible(False)
            self.more_info_button.setText("more information")
            self.adjustSize()
            self.showing_more_info = False
        else:
            self.text_edit.setVisible(True)
            self.more_info_button.setText("hide additional information")
            self.text_edit.setMinimumHeight(500)
            self.text_edit.setMinimumWidth(800)
            self.showing_more_info = True
        self.adjustSize()


def exception_hook(*exc_info):
    """Used to overwrite sys.excepthook, so that an exception does not
    terminate the program, but simply shows a Message with the exception.
    If the Exception is a KeyboardInterrupt, it does nothing, so that the
    interrupt may actually stop the program execution.

    Parameters
    ----------
    *exc_info : tuple(class, Exception, traceback)
        The information for the exception.
    """
    if issubclass(exc_info[0], KeyboardInterrupt):
        return
    elif issubclass(exc_info[0], RunEngineInterrupted):
        return
    if variables_handling.preferences["play_camel_on_error"]:
        try:
            effect = QSoundEffect()
            effect.setSource(
                QUrl.fromLocalFile(
                    str(resources.files(graphics) / "Camel-Groan-2-QuickSounds.com.wav")
                )
            )
            effect.play()
        except Exception as e:
            print(e)
    ErrorMessage(exc_info).exec()



"""Provides a function to rename typical names for fit-variables to valid
python-names"""

fit_variable_changer = {
    "+": "plus",
    "-": "minus",
    "=": "equals",
    "==": "equals",
    "/": "divide",
    "*": "times",
    "**": "power",
    "<": "less",
    ">": "greater",
    "(": "bracketL",
    ")": "bracketR",
    ".": "dot",
    ",": "comma",
    " ": "_",
}


def replace_name(var_name):
    """
    Replaces mathematical symbols with text so that the variable name becomes a
    valid name.

    Parameters
    ----------
    var_name : str
        The variable that should be renamed.
    """
    for key in fit_variable_changer:
        var_name = var_name.replace(key, fit_variable_changer[key])
    return var_name



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'mainWindow_v2.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QComboBox, QFrame,
    QGridLayout, QHBoxLayout, QHeaderView, QLabel,
    QLineEdit, QListWidgetItem, QMainWindow, QMenu,
    QMenuBar, QProgressBar, QPushButton, QScrollArea,
    QSizePolicy, QSpacerItem, QSplitter, QStatusBar,
    QTextEdit, QVBoxLayout, QWidget)

from nomad_camels.ui_widgets.console_redirect import Console_TextEdit
from nomad_camels.ui_widgets.run_queue import RunQueue
from nomad_camels.ui_widgets.variable_table import VariableTable

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.resize(1322, 610)
        self.actionPresets = QAction(MainWindow)
        self.actionPresets.setObjectName(u"actionPresets")
        self.actionOptions = QAction(MainWindow)
        self.actionOptions.setObjectName(u"actionOptions")
        self.actionSave_Device_Preset_As = QAction(MainWindow)
        self.actionSave_Device_Preset_As.setObjectName(u"actionSave_Device_Preset_As")
        self.actionLoad = QAction(MainWindow)
        self.actionLoad.setObjectName(u"actionLoad")
        self.actionMeasurement_Presets = QAction(MainWindow)
        self.actionMeasurement_Presets.setObjectName(u"actionMeasurement_Presets")
        self.actionSave_Preset = QAction(MainWindow)
        self.actionSave_Preset.setObjectName(u"actionSave_Preset")
        self.actionOpen_Backup_Device_Preset = QAction(MainWindow)
        self.actionOpen_Backup_Device_Preset.setObjectName(u"actionOpen_Backup_Device_Preset")
        self.actionLoad_Backup_Preset = QAction(MainWindow)
        self.actionLoad_Backup_Preset.setObjectName(u"actionLoad_Backup_Preset")
        self.actionAutosave_on_closing = QAction(MainWindow)
        self.actionAutosave_on_closing.setObjectName(u"actionAutosave_on_closing")
        self.actionAutosave_on_closing.setCheckable(True)
        self.actionUpdate_CAMELS = QAction(MainWindow)
        self.actionUpdate_CAMELS.setObjectName(u"actionUpdate_CAMELS")
        self.actionDark_Mode = QAction(MainWindow)
        self.actionDark_Mode.setObjectName(u"actionDark_Mode")
        self.actionDark_Mode.setCheckable(True)
        self.actionUndo = QAction(MainWindow)
        self.actionUndo.setObjectName(u"actionUndo")
        self.actionRedo = QAction(MainWindow)
        self.actionRedo.setObjectName(u"actionRedo")
        self.actionSettings = QAction(MainWindow)
        self.actionSettings.setObjectName(u"actionSettings")
        self.actionIOC_Builder = QAction(MainWindow)
        self.actionIOC_Builder.setObjectName(u"actionIOC_Builder")
        self.actionNew_Device_Preset_2 = QAction(MainWindow)
        self.actionNew_Device_Preset_2.setObjectName(u"actionNew_Device_Preset_2")
        self.actionSave_Device_Preset = QAction(MainWindow)
        self.actionSave_Device_Preset.setObjectName(u"actionSave_Device_Preset")
        self.actionSave_Preset_As = QAction(MainWindow)
        self.actionSave_Preset_As.setObjectName(u"actionSave_Preset_As")
        self.actionNew_Preset = QAction(MainWindow)
        self.actionNew_Preset.setObjectName(u"actionNew_Preset")
        self.actionNew_Device_Preset = QAction(MainWindow)
        self.actionNew_Device_Preset.setObjectName(u"actionNew_Device_Preset")
        self.actionDocumentation = QAction(MainWindow)
        self.actionDocumentation.setObjectName(u"actionDocumentation")
        self.actionReport_Bug = QAction(MainWindow)
        self.actionReport_Bug.setObjectName(u"actionReport_Bug")
        self.action_driver_builder = QAction(MainWindow)
        self.action_driver_builder.setObjectName(u"action_driver_builder")
        self.actionEPICS_driver_builder = QAction(MainWindow)
        self.actionEPICS_driver_builder.setObjectName(u"actionEPICS_driver_builder")
        self.actionExport_from_databroker = QAction(MainWindow)
        self.actionExport_from_databroker.setObjectName(u"actionExport_from_databroker")
        self.actionManage_Extensions = QAction(MainWindow)
        self.actionManage_Extensions.setObjectName(u"actionManage_Extensions")
        self.actionExport_CAMELS_hdf5_to_csv_json = QAction(MainWindow)
        self.actionExport_CAMELS_hdf5_to_csv_json.setObjectName(u"actionExport_CAMELS_hdf5_to_csv_json")
        self.actionQuit = QAction(MainWindow)
        self.actionQuit.setObjectName(u"actionQuit")
        self.actionWatchdogs = QAction(MainWindow)
        self.actionWatchdogs.setObjectName(u"actionWatchdogs")
        self.centralwidget = QWidget(MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.gridLayout_5 = QGridLayout(self.centralwidget)
        self.gridLayout_5.setObjectName(u"gridLayout_5")
        self.widget = QWidget(self.centralwidget)
        self.widget.setObjectName(u"widget")
        self.gridLayout_4 = QGridLayout(self.widget)
        self.gridLayout_4.setObjectName(u"gridLayout_4")
        self.gridLayout_4.setContentsMargins(0, 0, 0, 0)
        self.pushButton_resume = QPushButton(self.widget)
        self.pushButton_resume.setObjectName(u"pushButton_resume")
        self.pushButton_resume.setEnabled(False)
        self.pushButton_resume.setMaximumSize(QSize(130, 16777215))
        self.pushButton_resume.setStyleSheet(u"QPushButton {\n"
"                                background-color: #4CAF50; \n"
"                                color: white; \n"
"                                border: none; \n"
"                                padding: 2px 10px; \n"
"                                text-align: center; \n"
"                                text-decoration: none; \n"
"                                font-size: 13px; \n"
"                                margin: 2px 2px; \n"
"                                border-radius: 6px;\n"
"								font-weight: bold;\n"
"                            }\n"
"\n"
"                            QPushButton:hover {\n"
"                                background-color: #45a049;\n"
"                            }\n"
"QPushButton:disabled {\n"
"        background-color: #808080;\n"
"    }")

        self.gridLayout_4.addWidget(self.pushButton_resume, 1, 0, 1, 1)

        self.textEdit_console_output = Console_TextEdit(self.widget)
        self.textEdit_console_output.setObjectName(u"textEdit_console_output")
        self.textEdit_console_output.setMaximumSize(QSize(16777215, 16777215))
        self.textEdit_console_output.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByKeyboard|Qt.TextInteractionFlag.TextSelectableByMouse)

        self.gridLayout_4.addWidget(self.textEdit_console_output, 0, 4, 3, 1)

        self.pushButton_close_plots = QPushButton(self.widget)
        self.pushButton_close_plots.setObjectName(u"pushButton_close_plots")
        self.pushButton_close_plots.setStyleSheet(u"")

        self.gridLayout_4.addWidget(self.pushButton_close_plots, 3, 0, 1, 2)

        self.pushButton_stop = QPushButton(self.widget)
        self.pushButton_stop.setObjectName(u"pushButton_stop")
        self.pushButton_stop.setEnabled(False)
        self.pushButton_stop.setMaximumSize(QSize(130, 16777215))
        self.pushButton_stop.setStyleSheet(u"QPushButton {\n"
"                background-color: #E60000; \n"
"                color: white; \n"
"                                border: none; \n"
"                                padding: 2px 10px; \n"
"                                text-align: center; \n"
"                                text-decoration: none; \n"
"                                font-size: 13px; \n"
"                                margin: 2px 2px; \n"
"                                border-radius: 6px;\n"
"								font-weight: bold;\n"
"                            }\n"
"\n"
"            QPushButton:hover {\n"
"                background-color: #B22222;\n"
"            }\n"
"QPushButton:disabled {\n"
"        background-color: #808080;\n"
"    }")

        self.gridLayout_4.addWidget(self.pushButton_stop, 1, 2, 1, 1)

        self.pushButton_clear_log = QPushButton(self.widget)
        self.pushButton_clear_log.setObjectName(u"pushButton_clear_log")
        self.pushButton_clear_log.setStyleSheet(u"")

        self.gridLayout_4.addWidget(self.pushButton_clear_log, 3, 4, 1, 1)

        self.pushButton_pause = QPushButton(self.widget)
        self.pushButton_pause.setObjectName(u"pushButton_pause")
        self.pushButton_pause.setEnabled(False)
        self.pushButton_pause.setMaximumSize(QSize(130, 16777215))
        self.pushButton_pause.setStyleSheet(u"QPushButton {\n"
"        background-color: #FFA500;\n"
"        color: white;\n"
"        border: none;\n"
"        padding: 2px 10px;\n"
"        text-align: center;\n"
"        text-decoration: none;\n"
"        font-size: 13px;\n"
"        margin: 2px 2px;\n"
"        border-radius: 6px;\n"
"        font-weight: bold;\n"
"    }\n"
"\n"
"    QPushButton:hover {\n"
"        background-color: #FF8C00;\n"
"    }\n"
"QPushButton:disabled {\n"
"        background-color: #808080;\n"
"    }")

        self.gridLayout_4.addWidget(self.pushButton_pause, 1, 1, 1, 1)

        self.progressBar_protocols = QProgressBar(self.widget)
        self.progressBar_protocols.setObjectName(u"progressBar_protocols")
        self.progressBar_protocols.setMaximumSize(QSize(16777215, 16777215))
        self.progressBar_protocols.setValue(0)

        self.gridLayout_4.addWidget(self.progressBar_protocols, 2, 0, 1, 3)

        self.queue_variable_table = VariableTable(self.widget)
        self.queue_variable_table.setObjectName(u"queue_variable_table")

        self.gridLayout_4.addWidget(self.queue_variable_table, 0, 3, 4, 1)

        self.pushButton_show_log = QPushButton(self.widget)
        self.pushButton_show_log.setObjectName(u"pushButton_show_log")

        self.gridLayout_4.addWidget(self.pushButton_show_log, 3, 2, 1, 1)

        self.gridLayout_7 = QGridLayout()
        self.gridLayout_7.setObjectName(u"gridLayout_7")
        self.textEdit_meas_description = QTextEdit(self.widget)
        self.textEdit_meas_description.setObjectName(u"textEdit_meas_description")

        self.gridLayout_7.addWidget(self.textEdit_meas_description, 1, 0, 1, 1)

        self.label = QLabel(self.widget)
        self.label.setObjectName(u"label")
        font = QFont()
        font.setBold(True)
        self.label.setFont(font)

        self.gridLayout_7.addWidget(self.label, 0, 0, 1, 1)

        self.run_queue_widget = RunQueue(self.widget)
        self.run_queue_widget.setObjectName(u"run_queue_widget")

        self.gridLayout_7.addWidget(self.run_queue_widget, 1, 1, 1, 1)

        self.label_queue = QLabel(self.widget)
        self.label_queue.setObjectName(u"label_queue")
        self.label_queue.setFont(font)

        self.gridLayout_7.addWidget(self.label_queue, 0, 1, 1, 1)


        self.gridLayout_4.addLayout(self.gridLayout_7, 0, 0, 1, 3)


        self.gridLayout_5.addWidget(self.widget, 4, 9, 1, 2)

        self.label_logo = QLabel(self.centralwidget)
        self.label_logo.setObjectName(u"label_logo")
        self.label_logo.setMaximumSize(QSize(16777215, 70))
        self.label_logo.setPixmap(QPixmap(u"../graphics/camels_horizontal.png"))
        self.label_logo.setAlignment(Qt.AlignmentFlag.AlignRight|Qt.AlignmentFlag.AlignTrailing|Qt.AlignmentFlag.AlignVCenter)

        self.gridLayout_5.addWidget(self.label_logo, 5, 9, 1, 2)

        self.menu_widget = QWidget(self.centralwidget)
        self.menu_widget.setObjectName(u"menu_widget")
        self.menu_widget.setMaximumSize(QSize(16777215, 60))
        self.gridLayout_6 = QGridLayout(self.menu_widget)
        self.gridLayout_6.setObjectName(u"gridLayout_6")
        self.gridLayout_6.setContentsMargins(0, 0, 0, 0)
        self.pushButton_manage_instr = QPushButton(self.menu_widget)
        self.pushButton_manage_instr.setObjectName(u"pushButton_manage_instr")
        self.pushButton_manage_instr.setMaximumSize(QSize(150, 16777215))
        font1 = QFont()
        font1.setBold(True)
        font1.setUnderline(False)
        font1.setStrikeOut(False)
        self.pushButton_manage_instr.setFont(font1)
        self.pushButton_manage_instr.setStyleSheet(u"QPushButton {\n"
"        background-color: #2a4cdf;\n"
"        color: white;\n"
"        border: none;\n"
"        padding: 2px 10px;\n"
"        text-align: center;\n"
"        text-decoration: none;\n"
"        font-size: 15px;\n"
"        margin: 2px 2px;\n"
"        border-radius: 6px;\n"
"        font-weight: bold;\n"
"    }\n"
"\n"
"    QPushButton:hover {\n"
"        background-color: #1a3cbf;\n"
"    }")

        self.gridLayout_6.addWidget(self.pushButton_manage_instr, 0, 0, 1, 1)

        self.line_2 = QFrame(self.menu_widget)
        self.line_2.setObjectName(u"line_2")
        self.line_2.setFrameShape(QFrame.Shape.VLine)
        self.line_2.setFrameShadow(QFrame.Shadow.Sunken)

        self.gridLayout_6.addWidget(self.line_2, 0, 13, 1, 1)

        self.line = QFrame(self.menu_widget)
        self.line.setObjectName(u"line")
        self.line.setFrameShape(QFrame.Shape.VLine)
        self.line.setFrameShadow(QFrame.Shadow.Sunken)

        self.gridLayout_6.addWidget(self.line, 0, 8, 1, 1)

        self.comboBox_user_type = QComboBox(self.menu_widget)
        self.comboBox_user_type.setObjectName(u"comboBox_user_type")
        self.comboBox_user_type.setMaximumSize(QSize(110, 16777215))

        self.gridLayout_6.addWidget(self.comboBox_user_type, 0, 1, 1, 1)

        self.sample_widget = QWidget(self.menu_widget)
        self.sample_widget.setObjectName(u"sample_widget")
        self.verticalLayout_2 = QVBoxLayout(self.sample_widget)
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.sample_widget_default = QWidget(self.sample_widget)
        self.sample_widget_default.setObjectName(u"sample_widget_default")
        self.horizontalLayout_4 = QHBoxLayout(self.sample_widget_default)
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.horizontalLayout_4.setContentsMargins(0, 3, 0, 3)
        self.comboBox_sample = QComboBox(self.sample_widget_default)
        self.comboBox_sample.setObjectName(u"comboBox_sample")
        font2 = QFont()
        font2.setPointSize(10)
        self.comboBox_sample.setFont(font2)

        self.horizontalLayout_4.addWidget(self.comboBox_sample)

        self.pushButton_editSampleInfo = QPushButton(self.sample_widget_default)
        self.pushButton_editSampleInfo.setObjectName(u"pushButton_editSampleInfo")
        self.pushButton_editSampleInfo.setFont(font1)
        self.pushButton_editSampleInfo.setStyleSheet(u"QPushButton {\n"
"        background-color: #2a4cdf;\n"
"        color: white;\n"
"        border: none;\n"
"        padding: 2px 10px;\n"
"        text-align: center;\n"
"        text-decoration: none;\n"
"        font-size: 12px;\n"
"        margin: 2px 2px;\n"
"        border-radius: 6px;\n"
"        font-weight: bold;\n"
"    }\n"
"\n"
"    QPushButton:hover {\n"
"        background-color: #1a3cbf;\n"
"    }")

        self.horizontalLayout_4.addWidget(self.pushButton_editSampleInfo)


        self.verticalLayout_2.addWidget(self.sample_widget_default)

        self.sample_widget_nomad = QWidget(self.sample_widget)
        self.sample_widget_nomad.setObjectName(u"sample_widget_nomad")
        self.horizontalLayout_3 = QHBoxLayout(self.sample_widget_nomad)
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.horizontalLayout_3.setContentsMargins(0, 3, 0, 3)
        self.checkBox_use_nomad_sample = QCheckBox(self.sample_widget_nomad)
        self.checkBox_use_nomad_sample.setObjectName(u"checkBox_use_nomad_sample")

        self.horizontalLayout_3.addWidget(self.checkBox_use_nomad_sample)

        self.pushButton_nomad_sample = QPushButton(self.sample_widget_nomad)
        self.pushButton_nomad_sample.setObjectName(u"pushButton_nomad_sample")
        self.pushButton_nomad_sample.setFont(font1)
        self.pushButton_nomad_sample.setStyleSheet(u"QPushButton {\n"
"        background-color: #2a4cdf;\n"
"        color: white;\n"
"        border: none;\n"
"        padding: 2px 10px;\n"
"        text-align: center;\n"
"        text-decoration: none;\n"
"        font-size: 12px;\n"
"        margin: 2px 2px;\n"
"        border-radius: 6px;\n"
"        font-weight: bold;\n"
"    }\n"
"\n"
"    QPushButton:hover {\n"
"        background-color: #1a3cbf;\n"
"    }")

        self.horizontalLayout_3.addWidget(self.pushButton_nomad_sample)


        self.verticalLayout_2.addWidget(self.sample_widget_nomad)


        self.gridLayout_6.addWidget(self.sample_widget, 0, 10, 1, 1)

        self.session_upload_widget = QWidget(self.menu_widget)
        self.session_upload_widget.setObjectName(u"session_upload_widget")
        self.verticalLayout_3 = QVBoxLayout(self.session_upload_widget)
        self.verticalLayout_3.setSpacing(0)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.nomad_upload_widget = QWidget(self.session_upload_widget)
        self.nomad_upload_widget.setObjectName(u"nomad_upload_widget")
        self.horizontalLayout_6 = QHBoxLayout(self.nomad_upload_widget)
        self.horizontalLayout_6.setObjectName(u"horizontalLayout_6")
        self.horizontalLayout_6.setContentsMargins(0, 3, 0, 3)
        self.label_nomad_upload = QLabel(self.nomad_upload_widget)
        self.label_nomad_upload.setObjectName(u"label_nomad_upload")
        font3 = QFont()
        font3.setPointSize(10)
        font3.setBold(True)
        self.label_nomad_upload.setFont(font3)

        self.horizontalLayout_6.addWidget(self.label_nomad_upload)

        self.comboBox_upload_type = QComboBox(self.nomad_upload_widget)
        self.comboBox_upload_type.setObjectName(u"comboBox_upload_type")

        self.horizontalLayout_6.addWidget(self.comboBox_upload_type)

        self.comboBox_upload_choice = QComboBox(self.nomad_upload_widget)
        self.comboBox_upload_choice.setObjectName(u"comboBox_upload_choice")

        self.horizontalLayout_6.addWidget(self.comboBox_upload_choice)


        self.verticalLayout_3.addWidget(self.nomad_upload_widget)

        self.session_widget = QWidget(self.session_upload_widget)
        self.session_widget.setObjectName(u"session_widget")
        self.horizontalLayout_5 = QHBoxLayout(self.session_widget)
        self.horizontalLayout_5.setObjectName(u"horizontalLayout_5")
        self.horizontalLayout_5.setContentsMargins(0, 3, 0, 3)
        self.label_4 = QLabel(self.session_widget)
        self.label_4.setObjectName(u"label_4")
        self.label_4.setMaximumSize(QSize(1500000, 16777215))
        self.label_4.setFont(font3)
        self.label_4.setAlignment(Qt.AlignmentFlag.AlignRight|Qt.AlignmentFlag.AlignTrailing|Qt.AlignmentFlag.AlignVCenter)

        self.horizontalLayout_5.addWidget(self.label_4)

        self.lineEdit_session = QLineEdit(self.session_widget)
        self.lineEdit_session.setObjectName(u"lineEdit_session")

        self.horizontalLayout_5.addWidget(self.lineEdit_session)

        self.label_6 = QLabel(self.session_widget)
        self.label_6.setObjectName(u"label_6")
        self.label_6.setFont(font3)

        self.horizontalLayout_5.addWidget(self.label_6)

        self.lineEdit_tags = QLineEdit(self.session_widget)
        self.lineEdit_tags.setObjectName(u"lineEdit_tags")

        self.horizontalLayout_5.addWidget(self.lineEdit_tags)

        self.scrollArea = QScrollArea(self.session_widget)
        self.scrollArea.setObjectName(u"scrollArea")
        self.scrollArea.setWidgetResizable(True)
        self.scrollAreaWidgetContents = QWidget()
        self.scrollAreaWidgetContents.setObjectName(u"scrollAreaWidgetContents")
        self.scrollAreaWidgetContents.setGeometry(QRect(0, 0, 138, 24))
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)

        self.horizontalLayout_5.addWidget(self.scrollArea)


        self.verticalLayout_3.addWidget(self.session_widget)


        self.gridLayout_6.addWidget(self.session_upload_widget, 0, 15, 1, 1)

        self.user_widget = QWidget(self.menu_widget)
        self.user_widget.setObjectName(u"user_widget")
        self.verticalLayout = QVBoxLayout(self.user_widget)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.user_widget_default = QWidget(self.user_widget)
        self.user_widget_default.setObjectName(u"user_widget_default")
        self.horizontalLayout = QHBoxLayout(self.user_widget_default)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(0, 3, 0, 3)
        self.comboBox_user = QComboBox(self.user_widget_default)
        self.comboBox_user.setObjectName(u"comboBox_user")
        self.comboBox_user.setFont(font2)

        self.horizontalLayout.addWidget(self.comboBox_user)

        self.pushButton_editUserInfo = QPushButton(self.user_widget_default)
        self.pushButton_editUserInfo.setObjectName(u"pushButton_editUserInfo")
        self.pushButton_editUserInfo.setFont(font1)
        self.pushButton_editUserInfo.setStyleSheet(u"QPushButton {\n"
"        background-color: #2a4cdf;\n"
"        color: white;\n"
"        border: none;\n"
"        padding: 2px 10px;\n"
"        text-align: center;\n"
"        text-decoration: none;\n"
"        font-size: 12px;\n"
"        margin: 2px 2px;\n"
"        border-radius: 6px;\n"
"        font-weight: bold;\n"
"    }\n"
"\n"
"    QPushButton:hover {\n"
"        background-color: #1a3cbf;\n"
"    }")

        self.horizontalLayout.addWidget(self.pushButton_editUserInfo)


        self.verticalLayout.addWidget(self.user_widget_default)

        self.user_widget_nomad = QWidget(self.user_widget)
        self.user_widget_nomad.setObjectName(u"user_widget_nomad")
        self.horizontalLayout_2 = QHBoxLayout(self.user_widget_nomad)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(0, 3, 0, 3)
        self.label_nomad_user = QLabel(self.user_widget_nomad)
        self.label_nomad_user.setObjectName(u"label_nomad_user")
        self.label_nomad_user.setFont(font2)

        self.horizontalLayout_2.addWidget(self.label_nomad_user)

        self.pushButton_login_nomad = QPushButton(self.user_widget_nomad)
        self.pushButton_login_nomad.setObjectName(u"pushButton_login_nomad")
        self.pushButton_login_nomad.setFont(font1)
        self.pushButton_login_nomad.setStyleSheet(u"QPushButton {\n"
"        background-color: #2a4cdf;\n"
"        color: white;\n"
"        border: none;\n"
"        padding: 2px 10px;\n"
"        text-align: center;\n"
"        text-decoration: none;\n"
"        font-size: 12px;\n"
"        margin: 2px 2px;\n"
"        border-radius: 6px;\n"
"        font-weight: bold;\n"
"    }\n"
"\n"
"    QPushButton:hover {\n"
"        background-color: #1a3cbf;\n"
"    }")

        self.horizontalLayout_2.addWidget(self.pushButton_login_nomad)


        self.verticalLayout.addWidget(self.user_widget_nomad)


        self.gridLayout_6.addWidget(self.user_widget, 0, 2, 1, 1)

        self.label_8 = QLabel(self.menu_widget)
        self.label_8.setObjectName(u"label_8")
        self.label_8.setMaximumSize(QSize(70, 16777215))
        self.label_8.setFont(font3)
        self.label_8.setAlignment(Qt.AlignmentFlag.AlignRight|Qt.AlignmentFlag.AlignTrailing|Qt.AlignmentFlag.AlignVCenter)

        self.gridLayout_6.addWidget(self.label_8, 0, 9, 1, 1)


        self.gridLayout_5.addWidget(self.menu_widget, 1, 2, 1, 9)

        self.widget_2 = QWidget(self.centralwidget)
        self.widget_2.setObjectName(u"widget_2")
        self.gridLayout_3 = QGridLayout(self.widget_2)
        self.gridLayout_3.setSpacing(0)
        self.gridLayout_3.setObjectName(u"gridLayout_3")
        self.gridLayout_3.setContentsMargins(0, 0, 0, 0)
        self.label_no_instruments = QLabel(self.widget_2)
        self.label_no_instruments.setObjectName(u"label_no_instruments")

        self.gridLayout_3.addWidget(self.label_no_instruments, 0, 1, 1, 1)

        self.label_arrow = QLabel(self.widget_2)
        self.label_arrow.setObjectName(u"label_arrow")
        self.label_arrow.setMaximumSize(QSize(120, 16777215))
        self.label_arrow.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.gridLayout_3.addWidget(self.label_arrow, 0, 0, 1, 1)

        self.main_splitter = QSplitter(self.widget_2)
        self.main_splitter.setObjectName(u"main_splitter")
        self.main_splitter.setOrientation(Qt.Orientation.Vertical)
        self.manual_widget = QWidget(self.main_splitter)
        self.manual_widget.setObjectName(u"manual_widget")
        self.gridLayout = QGridLayout(self.manual_widget)
        self.gridLayout.setObjectName(u"gridLayout")
        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.gridLayout.addItem(self.horizontalSpacer, 1, 2, 1, 1)

        self.pushButton_add_manual = QPushButton(self.manual_widget)
        self.pushButton_add_manual.setObjectName(u"pushButton_add_manual")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pushButton_add_manual.sizePolicy().hasHeightForWidth())
        self.pushButton_add_manual.setSizePolicy(sizePolicy)
        self.pushButton_add_manual.setMinimumSize(QSize(32, 32))
        self.pushButton_add_manual.setMaximumSize(QSize(32, 32))
        self.pushButton_add_manual.setFont(font1)
        self.pushButton_add_manual.setStyleSheet(u"QPushButton {\n"
"                                background-color: #4CAF50; \n"
"                                color: white; \n"
"                                border: none; \n"
"                                padding: 0px; \n"
"                                padding-bottom: 5px;\n"
"                                text-align: center; \n"
"                                text-decoration: none; \n"
"                                font-size: 30px; \n"
"                                margin: 2px 2px; \n"
"                                border-radius: 6px;\n"
"								font-weight: bold;\n"
"                            }\n"
"\n"
"                            QPushButton:hover {\n"
"                                background-color: #45a049;\n"
"                            }")

        self.gridLayout.addWidget(self.pushButton_add_manual, 1, 1, 1, 1)

        self.label_2 = QLabel(self.manual_widget)
        self.label_2.setObjectName(u"label_2")
        font4 = QFont()
        font4.setFamilies([u"Calibri"])
        font4.setPointSize(20)
        font4.setBold(True)
        self.label_2.setFont(font4)
        self.label_2.setStyleSheet(u"QLabel {\n"
"	font-family: Calibri;\n"
"	font-size: 20pt;\n"
"	font-weight: bold;\n"
"}")

        self.gridLayout.addWidget(self.label_2, 1, 0, 1, 1)

        self.main_splitter.addWidget(self.manual_widget)
        self.meas_widget = QWidget(self.main_splitter)
        self.meas_widget.setObjectName(u"meas_widget")
        self.gridLayout_2 = QGridLayout(self.meas_widget)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.gridLayout_2.addItem(self.horizontalSpacer_2, 1, 3, 1, 1)

        self.label_3 = QLabel(self.meas_widget)
        self.label_3.setObjectName(u"label_3")
        self.label_3.setFont(font4)
        self.label_3.setStyleSheet(u"QLabel {\n"
"	font-family: Calibri;\n"
"	font-size: 20pt;\n"
"	font-weight: bold;\n"
"}")

        self.gridLayout_2.addWidget(self.label_3, 1, 0, 1, 1)

        self.pushButton_add_meas = QPushButton(self.meas_widget)
        self.pushButton_add_meas.setObjectName(u"pushButton_add_meas")
        sizePolicy.setHeightForWidth(self.pushButton_add_meas.sizePolicy().hasHeightForWidth())
        self.pushButton_add_meas.setSizePolicy(sizePolicy)
        self.pushButton_add_meas.setMinimumSize(QSize(32, 32))
        self.pushButton_add_meas.setMaximumSize(QSize(32, 32))
        self.pushButton_add_meas.setFont(font1)
        self.pushButton_add_meas.setStyleSheet(u"QPushButton {\n"
"                                background-color: #4CAF50; \n"
"                                color: white; \n"
"                                border: none; \n"
"                                padding: 0px; \n"
"                                padding-bottom: 5px;\n"
"                                text-align: center; \n"
"                                text-decoration: none; \n"
"                                font-size: 30px; \n"
"                                margin: 2px 2px; \n"
"                                border-radius: 6px;\n"
"								font-weight: bold;\n"
"                            }\n"
"\n"
"                            QPushButton:hover {\n"
"                                background-color: #45a049;\n"
"                            }")

        self.gridLayout_2.addWidget(self.pushButton_add_meas, 1, 1, 1, 1)

        self.pushButton_import_protocol = QPushButton(self.meas_widget)
        self.pushButton_import_protocol.setObjectName(u"pushButton_import_protocol")
        sizePolicy.setHeightForWidth(self.pushButton_import_protocol.sizePolicy().hasHeightForWidth())
        self.pushButton_import_protocol.setSizePolicy(sizePolicy)
        self.pushButton_import_protocol.setMinimumSize(QSize(32, 32))
        self.pushButton_import_protocol.setMaximumSize(QSize(60, 32))
        self.pushButton_import_protocol.setStyleSheet(u"QPushButton {\n"
"        background-color: #2a4cdf;\n"
"        color: white;\n"
"        border: none;\n"
"        padding: 2px 10px;\n"
"        text-align: center;\n"
"        text-decoration: none;\n"
"        font-size: 12px;\n"
"        margin: 2px 2px;\n"
"        border-radius: 6px;\n"
"        font-weight: bold;\n"
"    }\n"
"\n"
"    QPushButton:hover {\n"
"        background-color: #1a3cbf;\n"
"    }")

        self.gridLayout_2.addWidget(self.pushButton_import_protocol, 1, 2, 1, 1)

        self.main_splitter.addWidget(self.meas_widget)

        self.gridLayout_3.addWidget(self.main_splitter, 1, 0, 1, 2)


        self.gridLayout_5.addWidget(self.widget_2, 4, 2, 2, 7)

        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(MainWindow)
        self.menubar.setObjectName(u"menubar")
        self.menubar.setGeometry(QRect(0, 0, 1322, 33))
        self.menuFile = QMenu(self.menubar)
        self.menuFile.setObjectName(u"menuFile")
        self.menuHelp = QMenu(self.menubar)
        self.menuHelp.setObjectName(u"menuHelp")
        self.menuTools = QMenu(self.menubar)
        self.menuTools.setObjectName(u"menuTools")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QStatusBar(MainWindow)
        self.statusbar.setObjectName(u"statusbar")
        MainWindow.setStatusBar(self.statusbar)

        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuTools.menuAction())
        self.menubar.addAction(self.menuHelp.menuAction())
        self.menuFile.addAction(self.actionNew_Preset)
        self.menuFile.addAction(self.actionSave_Preset)
        self.menuFile.addAction(self.actionSave_Preset_As)
        self.menuFile.addAction(self.actionLoad_Backup_Preset)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionSettings)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionQuit)
        self.menuHelp.addAction(self.actionDocumentation)
        self.menuHelp.addAction(self.actionReport_Bug)
        self.menuTools.addAction(self.actionWatchdogs)
        self.menuTools.addSeparator()
        self.menuTools.addAction(self.actionUpdate_CAMELS)
        self.menuTools.addSeparator()
        self.menuTools.addAction(self.actionExport_from_databroker)
        self.menuTools.addAction(self.actionExport_CAMELS_hdf5_to_csv_json)
        self.menuTools.addSeparator()
        self.menuTools.addAction(self.action_driver_builder)
        self.menuTools.addAction(self.actionEPICS_driver_builder)
        self.menuTools.addSeparator()
        self.menuTools.addAction(self.actionManage_Extensions)

        self.retranslateUi(MainWindow)

        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"MainWindow", None))
        self.actionPresets.setText(QCoreApplication.translate("MainWindow", u"Device-Presets", None))
        self.actionOptions.setText(QCoreApplication.translate("MainWindow", u"Options", None))
        self.actionSave_Device_Preset_As.setText(QCoreApplication.translate("MainWindow", u"Save Device Preset As", None))
        self.actionLoad.setText(QCoreApplication.translate("MainWindow", u"Open", None))
        self.actionMeasurement_Presets.setText(QCoreApplication.translate("MainWindow", u"Measurement-Presets", None))
        self.actionSave_Preset.setText(QCoreApplication.translate("MainWindow", u"Save Preset", None))
        self.actionOpen_Backup_Device_Preset.setText(QCoreApplication.translate("MainWindow", u"Load Backup Device Preset", None))
        self.actionLoad_Backup_Preset.setText(QCoreApplication.translate("MainWindow", u"Load Preset", None))
        self.actionAutosave_on_closing.setText(QCoreApplication.translate("MainWindow", u"Autosave on closing", None))
        self.actionUpdate_CAMELS.setText(QCoreApplication.translate("MainWindow", u"Update NOMAD-CAMELS", None))
        self.actionDark_Mode.setText(QCoreApplication.translate("MainWindow", u"Dark Mode", None))
        self.actionUndo.setText(QCoreApplication.translate("MainWindow", u"Undo (ctrl + z)", None))
        self.actionRedo.setText(QCoreApplication.translate("MainWindow", u"Redo (ctrl + y)", None))
        self.actionSettings.setText(QCoreApplication.translate("MainWindow", u"Settings", None))
        self.actionIOC_Builder.setText(QCoreApplication.translate("MainWindow", u"IOC-Builder", None))
        self.actionNew_Device_Preset_2.setText(QCoreApplication.translate("MainWindow", u"New Device Preset", None))
        self.actionSave_Device_Preset.setText(QCoreApplication.translate("MainWindow", u"Save Device Preset", None))
        self.actionSave_Preset_As.setText(QCoreApplication.translate("MainWindow", u"Save Preset As", None))
        self.actionNew_Preset.setText(QCoreApplication.translate("MainWindow", u"New Preset", None))
        self.actionNew_Device_Preset.setText(QCoreApplication.translate("MainWindow", u"New Device Preset", None))
        self.actionDocumentation.setText(QCoreApplication.translate("MainWindow", u"Documentation", None))
        self.actionReport_Bug.setText(QCoreApplication.translate("MainWindow", u"Report Bug", None))
        self.action_driver_builder.setText(QCoreApplication.translate("MainWindow", u"Driver builder", None))
        self.actionEPICS_driver_builder.setText(QCoreApplication.translate("MainWindow", u"EPICS-driver-builder", None))
        self.actionExport_from_databroker.setText(QCoreApplication.translate("MainWindow", u"Export from databroker", None))
        self.actionManage_Extensions.setText(QCoreApplication.translate("MainWindow", u"Manage Extensions", None))
        self.actionExport_CAMELS_hdf5_to_csv_json.setText(QCoreApplication.translate("MainWindow", u"Export CAMELS hdf5 to csv/json", None))
        self.actionQuit.setText(QCoreApplication.translate("MainWindow", u"Quit", None))
        self.actionWatchdogs.setText(QCoreApplication.translate("MainWindow", u"Watchdogs", None))
        self.pushButton_resume.setText(QCoreApplication.translate("MainWindow", u"Resume", None))
#if QT_CONFIG(tooltip)
        self.pushButton_close_plots.setToolTip(QCoreApplication.translate("MainWindow", u"close all open plots\n"
"plots from a running protocol reopen with new data", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton_close_plots.setText(QCoreApplication.translate("MainWindow", u"Close Plots", None))
#if QT_CONFIG(tooltip)
        self.pushButton_stop.setToolTip(QCoreApplication.translate("MainWindow", u"abort the running protocol", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton_stop.setText(QCoreApplication.translate("MainWindow", u"Stop", None))
        self.pushButton_clear_log.setText(QCoreApplication.translate("MainWindow", u"Clear Log", None))
        self.pushButton_pause.setText(QCoreApplication.translate("MainWindow", u"Pause", None))
        self.pushButton_show_log.setText(QCoreApplication.translate("MainWindow", u"Show Log", None))
        self.label.setText(QCoreApplication.translate("MainWindow", u"Measurement Description", None))
        self.label_queue.setText(QCoreApplication.translate("MainWindow", u"Queue", None))
        self.label_logo.setText("")
        self.pushButton_manage_instr.setText(QCoreApplication.translate("MainWindow", u"Manage\n"
"Instruments", None))
        self.pushButton_editSampleInfo.setText(QCoreApplication.translate("MainWindow", u"Edit Sample-Information", None))
#if QT_CONFIG(tooltip)
        self.checkBox_use_nomad_sample.setToolTip(QCoreApplication.translate("MainWindow", u"only when this box is checked, the sample from NOMAD is used for the data", None))
#endif // QT_CONFIG(tooltip)
        self.checkBox_use_nomad_sample.setText(QCoreApplication.translate("MainWindow", u"use NOMAD sample", None))
        self.pushButton_nomad_sample.setText(QCoreApplication.translate("MainWindow", u"select NOMAD sample", None))
        self.label_nomad_upload.setText(QCoreApplication.translate("MainWindow", u"NOMAD Upload:", None))
        self.label_4.setText(QCoreApplication.translate("MainWindow", u"Session:", None))
#if QT_CONFIG(tooltip)
        self.lineEdit_session.setToolTip(QCoreApplication.translate("MainWindow", u"used to distinguish measurements\n"
"adds a sub-folder for the data", None))
#endif // QT_CONFIG(tooltip)
        self.label_6.setText(QCoreApplication.translate("MainWindow", u"Tags", None))
#if QT_CONFIG(tooltip)
        self.lineEdit_tags.setToolTip(QCoreApplication.translate("MainWindow", u"useful to identify / search measurements", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton_editUserInfo.setText(QCoreApplication.translate("MainWindow", u"Edit User-Information", None))
        self.label_nomad_user.setText(QCoreApplication.translate("MainWindow", u"not logged in", None))
        self.pushButton_login_nomad.setText(QCoreApplication.translate("MainWindow", u"NOMAD login", None))
        self.label_8.setText(QCoreApplication.translate("MainWindow", u"Sample:", None))
        self.label_no_instruments.setText(QCoreApplication.translate("MainWindow", u"<html><head/><body><p><span style=\" font-size:12pt; font-weight:600;\">You are currently using no instruments.</span></p><p><span style=\" font-size:12pt; font-weight:600;\">Click &quot;Manage Instruments&quot; to configure your</span></p><p><span style=\" font-size:12pt; font-weight:600;\">first instrument and start with NOMAD CAMELS!</span></p></body></html>", None))
        self.label_arrow.setText("")
#if QT_CONFIG(tooltip)
        self.pushButton_add_manual.setToolTip(QCoreApplication.translate("MainWindow", u"add a new manual control", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton_add_manual.setText(QCoreApplication.translate("MainWindow", u"+", None))
        self.label_2.setText(QCoreApplication.translate("MainWindow", u"Manual Control", None))
        self.label_3.setText(QCoreApplication.translate("MainWindow", u"Measurement Protocols", None))
#if QT_CONFIG(tooltip)
        self.pushButton_add_meas.setToolTip(QCoreApplication.translate("MainWindow", u"create a new protocol", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton_add_meas.setText(QCoreApplication.translate("MainWindow", u"+", None))
        self.pushButton_import_protocol.setText(QCoreApplication.translate("MainWindow", u"import", None))
        self.menuFile.setTitle(QCoreApplication.translate("MainWindow", u"File", None))
        self.menuHelp.setTitle(QCoreApplication.translate("MainWindow", u"Help", None))
        self.menuTools.setTitle(QCoreApplication.translate("MainWindow", u"Tools", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ui_stage_control.ui'
##
## Created by: Qt User Interface Compiler version 6.5.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QCheckBox,
    QFrame,
    QGridLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QWidget,
)


class Ui_Form(object):
    """ """

    def setupUi(self, Form):
        """

        Parameters
        ----------
        Form :


        Returns
        -------

        """
        if not Form.objectName():
            Form.setObjectName("Form")
        Form.resize(424, 279)
        self.gridLayout_2 = QGridLayout(Form)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.horizontalSpacer_4 = QSpacerItem(
            0, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.gridLayout_2.addItem(self.horizontalSpacer_4, 3, 5, 1, 3)

        self.pushButton_right = QPushButton(Form)
        self.pushButton_right.setObjectName("pushButton_right")

        self.gridLayout_2.addWidget(self.pushButton_right, 3, 4, 1, 1)

        self.pushButton_up = QPushButton(Form)
        self.pushButton_up.setObjectName("pushButton_up")

        self.gridLayout_2.addWidget(self.pushButton_up, 2, 3, 1, 1)

        self.pushButton_down = QPushButton(Form)
        self.pushButton_down.setObjectName("pushButton_down")

        self.gridLayout_2.addWidget(self.pushButton_down, 4, 3, 1, 1)

        self.pushButton_stop = QPushButton(Form)
        self.pushButton_stop.setObjectName("pushButton_stop")

        self.gridLayout_2.addWidget(self.pushButton_stop, 3, 3, 1, 1)

        self.pushButton_left = QPushButton(Form)
        self.pushButton_left.setObjectName("pushButton_left")

        self.gridLayout_2.addWidget(self.pushButton_left, 3, 2, 1, 1)

        self.lineEdit_stepZ = QLineEdit(Form)
        self.lineEdit_stepZ.setObjectName("lineEdit_stepZ")

        self.gridLayout_2.addWidget(self.lineEdit_stepZ, 4, 10, 1, 1)

        self.label = QLabel(Form)
        self.label.setObjectName("label")
        self.label.setLayoutDirection(Qt.LeftToRight)
        self.label.setAlignment(Qt.AlignRight | Qt.AlignTrailing | Qt.AlignVCenter)

        self.gridLayout_2.addWidget(self.label, 2, 8, 1, 1)

        self.lineEdit_stepY = QLineEdit(Form)
        self.lineEdit_stepY.setObjectName("lineEdit_stepY")

        self.gridLayout_2.addWidget(self.lineEdit_stepY, 3, 10, 1, 1)

        self.pushButton_zDown = QPushButton(Form)
        self.pushButton_zDown.setObjectName("pushButton_zDown")

        self.gridLayout_2.addWidget(self.pushButton_zDown, 4, 6, 1, 1)

        self.pushButton_zUp = QPushButton(Form)
        self.pushButton_zUp.setObjectName("pushButton_zUp")

        self.gridLayout_2.addWidget(self.pushButton_zUp, 2, 6, 1, 1)

        self.horizontalSpacer_2 = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.gridLayout_2.addItem(self.horizontalSpacer_2, 3, 0, 1, 1)

        self.horizontalSpacer_3 = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.gridLayout_2.addItem(self.horizontalSpacer_3, 2, 7, 1, 1)

        self.horizontalSpacer = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.gridLayout_2.addItem(self.horizontalSpacer, 2, 0, 1, 3)

        self.lineEdit_goZ = QLineEdit(Form)
        self.lineEdit_goZ.setObjectName("lineEdit_goZ")

        self.gridLayout_2.addWidget(self.lineEdit_goZ, 10, 8, 1, 3)

        self.pushButton_go_to = QPushButton(Form)
        self.pushButton_go_to.setObjectName("pushButton_go_to")

        self.gridLayout_2.addWidget(self.pushButton_go_to, 7, 8, 1, 3)

        self.lineEdit_goX = QLineEdit(Form)
        self.lineEdit_goX.setObjectName("lineEdit_goX")

        self.gridLayout_2.addWidget(self.lineEdit_goX, 8, 8, 1, 3)

        self.label_2 = QLabel(Form)
        self.label_2.setObjectName("label_2")
        self.label_2.setAlignment(Qt.AlignRight | Qt.AlignTrailing | Qt.AlignVCenter)

        self.gridLayout_2.addWidget(self.label_2, 3, 8, 1, 1)

        self.label_3 = QLabel(Form)
        self.label_3.setObjectName("label_3")
        self.label_3.setAlignment(Qt.AlignRight | Qt.AlignTrailing | Qt.AlignVCenter)

        self.gridLayout_2.addWidget(self.label_3, 4, 8, 1, 1)

        self.lineEdit_goY = QLineEdit(Form)
        self.lineEdit_goY.setObjectName("lineEdit_goY")

        self.gridLayout_2.addWidget(self.lineEdit_goY, 9, 8, 1, 3)

        self.verticalSpacer = QSpacerItem(
            20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding
        )

        self.gridLayout_2.addItem(self.verticalSpacer, 0, 10, 1, 1)

        self.line = QFrame(Form)
        self.line.setObjectName("line")
        self.line.setFrameShape(QFrame.VLine)
        self.line.setFrameShadow(QFrame.Sunken)

        self.gridLayout_2.addWidget(self.line, 1, 11, 4, 1)

        self.lineEdit_currentX = QLineEdit(Form)
        self.lineEdit_currentX.setObjectName("lineEdit_currentX")
        self.lineEdit_currentX.setEnabled(False)

        self.gridLayout_2.addWidget(self.lineEdit_currentX, 8, 13, 1, 2)

        self.lineEdit_currentZ = QLineEdit(Form)
        self.lineEdit_currentZ.setObjectName("lineEdit_currentZ")
        self.lineEdit_currentZ.setEnabled(False)

        self.gridLayout_2.addWidget(self.lineEdit_currentZ, 10, 13, 1, 2)

        self.lineEdit_stepX = QLineEdit(Form)
        self.lineEdit_stepX.setObjectName("lineEdit_stepX")

        self.gridLayout_2.addWidget(self.lineEdit_stepX, 2, 10, 1, 1)

        self.label_6 = QLabel(Form)
        self.label_6.setObjectName("label_6")

        self.gridLayout_2.addWidget(self.label_6, 7, 13, 1, 2)

        self.label_4 = QLabel(Form)
        self.label_4.setObjectName("label_4")
        font = QFont()
        font.setBold(True)
        self.label_4.setFont(font)

        self.gridLayout_2.addWidget(self.label_4, 1, 10, 1, 1)

        self.lineEdit_currentY = QLineEdit(Form)
        self.lineEdit_currentY.setObjectName("lineEdit_currentY")
        self.lineEdit_currentY.setEnabled(False)

        self.gridLayout_2.addWidget(self.lineEdit_currentY, 9, 13, 1, 2)

        self.horizontalSpacer_7 = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.gridLayout_2.addItem(self.horizontalSpacer_7, 4, 0, 1, 3)

        self.checkBox_manualActive = QCheckBox(Form)
        self.checkBox_manualActive.setObjectName("checkBox_manualActive")

        self.gridLayout_2.addWidget(self.checkBox_manualActive, 11, 2, 1, 5)

        self.label_5 = QLabel(Form)
        self.label_5.setObjectName("label_5")
        self.label_5.setFont(font)

        self.gridLayout_2.addWidget(self.label_5, 7, 2, 1, 5)

        self.lineEdit_manualY = QLineEdit(Form)
        self.lineEdit_manualY.setObjectName("lineEdit_manualY")

        self.gridLayout_2.addWidget(self.lineEdit_manualY, 9, 2, 1, 5)

        self.horizontalSpacer_6 = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.gridLayout_2.addItem(self.horizontalSpacer_6, 3, 1, 1, 1)

        self.lineEdit_manualZ = QLineEdit(Form)
        self.lineEdit_manualZ.setObjectName("lineEdit_manualZ")

        self.gridLayout_2.addWidget(self.lineEdit_manualZ, 10, 2, 1, 5)

        self.line_3 = QFrame(Form)
        self.line_3.setObjectName("line_3")
        self.line_3.setFrameShape(QFrame.VLine)
        self.line_3.setFrameShadow(QFrame.Sunken)

        self.gridLayout_2.addWidget(self.line_3, 7, 7, 5, 1)

        self.horizontalSpacer_5 = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.gridLayout_2.addItem(self.horizontalSpacer_5, 2, 4, 1, 2)

        self.lineEdit_manualX = QLineEdit(Form)
        self.lineEdit_manualX.setObjectName("lineEdit_manualX")

        self.gridLayout_2.addWidget(self.lineEdit_manualX, 8, 2, 1, 5)

        self.checkBox_refY = QCheckBox(Form)
        self.checkBox_refY.setObjectName("checkBox_refY")
        self.checkBox_refY.setChecked(True)

        self.gridLayout_2.addWidget(self.checkBox_refY, 3, 12, 1, 1)

        self.checkBox_refZ = QCheckBox(Form)
        self.checkBox_refZ.setObjectName("checkBox_refZ")
        self.checkBox_refZ.setChecked(True)

        self.gridLayout_2.addWidget(self.checkBox_refZ, 4, 12, 1, 1)

        self.checkBox_refX = QCheckBox(Form)
        self.checkBox_refX.setObjectName("checkBox_refX")
        self.checkBox_refX.setChecked(True)

        self.gridLayout_2.addWidget(self.checkBox_refX, 2, 12, 1, 1)

        self.pushButton_ref = QPushButton(Form)
        self.pushButton_ref.setObjectName("pushButton_ref")

        self.gridLayout_2.addWidget(self.pushButton_ref, 1, 12, 1, 3)

        self.pushButton_position = QPushButton(Form)
        self.pushButton_position.setObjectName("pushButton_position")

        self.gridLayout_2.addWidget(self.pushButton_position, 9, 11, 1, 2)

        self.lineEdit_read_frequ = QLineEdit(Form)
        self.lineEdit_read_frequ.setObjectName("lineEdit_read_frequ")

        self.gridLayout_2.addWidget(self.lineEdit_read_frequ, 11, 13, 1, 2)

        self.label_7 = QLabel(Form)
        self.label_7.setObjectName("label_7")
        self.label_7.setAlignment(Qt.AlignRight | Qt.AlignTrailing | Qt.AlignVCenter)

        self.gridLayout_2.addWidget(self.label_7, 11, 8, 1, 5)

        self.line_2 = QFrame(Form)
        self.line_2.setObjectName("line_2")
        self.line_2.setFrameShape(QFrame.HLine)
        self.line_2.setFrameShadow(QFrame.Sunken)

        self.gridLayout_2.addWidget(self.line_2, 5, 0, 1, 15)

        self.retranslateUi(Form)

        QMetaObject.connectSlotsByName(Form)

    # setupUi

    def retranslateUi(self, Form):
        """

        Parameters
        ----------
        Form :


        Returns
        -------

        """
        Form.setWindowTitle(QCoreApplication.translate("Form", "Form", None))
        self.pushButton_right.setText("")
        self.pushButton_up.setText("")
        self.pushButton_down.setText("")
        self.pushButton_stop.setText("")
        self.pushButton_left.setText("")
        self.label.setText(QCoreApplication.translate("Form", "x", None))
        self.pushButton_zDown.setText(QCoreApplication.translate("Form", "z", None))
        self.pushButton_zUp.setText(QCoreApplication.translate("Form", "z", None))
        self.pushButton_go_to.setText(QCoreApplication.translate("Form", "Go To", None))
        self.label_2.setText(QCoreApplication.translate("Form", "y", None))
        self.label_3.setText(QCoreApplication.translate("Form", "z", None))
        self.label_6.setText(QCoreApplication.translate("Form", "current pos.", None))
        self.label_4.setText(QCoreApplication.translate("Form", "Step Size", None))
        self.checkBox_manualActive.setText(
            QCoreApplication.translate("Form", "active", None)
        )
        self.label_5.setText(
            QCoreApplication.translate("Form", "Manual Movement Speeds", None)
        )
        self.checkBox_refY.setText(QCoreApplication.translate("Form", "y", None))
        self.checkBox_refZ.setText(QCoreApplication.translate("Form", "z", None))
        self.checkBox_refX.setText(QCoreApplication.translate("Form", "x", None))
        self.pushButton_ref.setText(
            QCoreApplication.translate("Form", "Find Reference", None)
        )
        self.pushButton_position.setText("")
        self.label_7.setText(
            QCoreApplication.translate("Form", "Time between readings (s)", None)
        )

    # retranslateUi



from PySide6.QtWidgets import QDialog, QStyleFactory, QMessageBox, QApplication
from PySide6.QtCore import Qt, QCoreApplication
from PySide6.QtGui import QKeyEvent
import qt_material

from nomad_camels.gui.settings_window import Ui_settings_window
from nomad_camels.utility import load_save_functions
from nomad_camels.utility.theme_changing import change_theme
from nomad_camels.utility.logging_settings import log_levels
from nomad_camels.nomad_integration.nomad_communication import make_correct_url

import secrets
import hashlib
import sqlite3
import os


class Settings_Window(Ui_settings_window, QDialog):
    """Dialog to change the settings used in CAMELS."""

    def __init__(self, parent=None, settings=None):
        super().__init__(parent)
        self.setupUi(self)
        self.setWindowTitle("Settings - NOMAD CAMELS")

        self.pathButton_config_path.set_path(load_save_functions.appdata_path)
        self.pathButton_config_path.select_directory = True

        themes = QStyleFactory.keys()
        themes.append("qt-material")
        self.comboBox_theme.addItems(themes)
        if "graphic_theme" in settings and settings["graphic_theme"] in themes:
            self.comboBox_theme.setCurrentText(settings["graphic_theme"])
        else:
            app = QCoreApplication.instance()
            self.comboBox_theme.setCurrentText(app.style().objectName())
        material_themes = []
        for t in qt_material.list_themes():
            if t.startswith("light_"):
                material_themes.append(t[6:-4])
        self.comboBox_material_theme.addItems(material_themes)
        if (
            "material_theme" in settings
            and settings["material_theme"] in material_themes
        ):
            self.comboBox_material_theme.setCurrentText(settings["material_theme"])
        if "dark_mode" in settings:
            self.checkBox_dark.setChecked(settings["dark_mode"])
        self.comboBox_theme.currentTextChanged.connect(self.change_theme)
        self.comboBox_material_theme.currentTextChanged.connect(self.change_theme)
        self.checkBox_dark.clicked.connect(self.change_theme)
        self.change_theme()

        self.pushButton_generate_Api_key.clicked.connect(self.generate_api_key)
        self.pushButton_copy_Api_key_clipboard.clicked.connect(self.copy_to_clipboard)
        self.pushButton_delete_Api_keys.clicked.connect(self.confirm_delete_api_keys)
        self.checkBox_enable_Api.clicked.connect(self.change_enable_API)

        standard_pref = load_save_functions.standard_pref

        if "autosave" in settings:
            self.checkBox_autosave.setChecked(settings["autosave"])
        else:
            self.checkBox_autosave.setChecked(standard_pref["autosave"])
        if "autosave_run" in settings:
            self.checkBox_autosave_run.setChecked(settings["autosave_run"])
        else:
            self.checkBox_autosave_run.setChecked(standard_pref["autosave_run"])
        if "backup_before_run" in settings:
            self.checkBox_backup_before_run.setChecked(settings["backup_before_run"])
        else:
            self.checkBox_backup_before_run.setChecked(
                standard_pref["backup_before_run"]
            )
        if "auto_check_updates" in settings:
            self.checkBox_auto_check_updates.setChecked(settings["auto_check_updates"])
        else:
            self.checkBox_auto_check_updates.setChecked(
                standard_pref["auto_check_updates"]
            )
        if "number_format" in settings:
            if settings["number_format"] == "plain":
                self.radioButton_plain_numbers.setChecked(True)
            elif settings["number_format"] == "scientific":
                self.radioButton_scientific.setChecked(True)
            else:
                self.radioButton_mixed.setChecked(True)
        else:
            self.radioButton_mixed.setChecked(True)
        if "mixed_from" in settings:
            self.spinBox_scientific_from.setValue(settings["mixed_from"])
        else:
            self.spinBox_scientific_from.setValue(standard_pref["mixed_from"])
        if "n_decimals" in settings:
            self.spinBox_n_decimals.setValue(settings["n_decimals"])
        else:
            self.spinBox_n_decimals.setValue(standard_pref["n_decimals"])
        if "py_files_path" in settings:
            self.pathButton_py_files.set_path(settings["py_files_path"])
        else:
            self.pathButton_py_files.set_path(standard_pref["py_files_path"])
        if "meas_files_path" in settings:
            self.pathButton_meas_files.set_path(settings["meas_files_path"])
        else:
            self.pathButton_meas_files.set_path(standard_pref["meas_files_path"])
        if "device_driver_path" in settings:
            self.pathButton_device_path.set_path(settings["device_driver_path"])
        else:
            self.pathButton_device_path.set_path(standard_pref["device_driver_path"])
        self.pathButton_py_files.select_directory = True
        self.pathButton_meas_files.select_directory = True
        self.pathButton_device_path.select_directory = True
        self.pathButton_extension_path.select_directory = True
        if "databroker_catalog_name" in settings:
            self.lineEdit_catalog_name.setText(settings["databroker_catalog_name"])
        else:
            self.lineEdit_catalog_name.setText(standard_pref["databroker_catalog_name"])
        if "number_databroker_files" in settings:
            self.spinBox_n_databroker_files.setValue(
                settings["number_databroker_files"]
            )
        else:
            self.spinBox_n_databroker_files.setValue(
                standard_pref["number_databroker_files"]
            )
        self.spinBox_n_databroker_files.setToolTip("0 files means keep all")
        if "play_camel_on_error" in settings:
            self.checkBox_play_camel_on_error.setChecked(
                settings["play_camel_on_error"]
            )
        else:
            self.checkBox_play_camel_on_error.setChecked(
                standard_pref["play_camel_on_error"]
            )

        for level in log_levels:
            self.comboBox_log_level.addItem(level)
        if "log_level" in settings:
            self.comboBox_log_level.setCurrentText(settings["log_level"])
        else:
            self.comboBox_log_level.setCurrentText(standard_pref["log_level"])
        if "logfile_size" in settings:
            self.spinBox_logfile_size.setValue(settings["logfile_size"])
        else:
            self.spinBox_logfile_size.setValue(standard_pref["logfile_size"])
        if "logfile_backups" in settings:
            self.spinBox_logfile_number.setValue(settings["logfile_backups"])
        else:
            self.spinBox_logfile_number.setValue(standard_pref["logfile_backups"])
        if "backups" in settings:
            if settings["backups"] == "all":
                self.radioButton_all_backups.setChecked(True)
            elif settings["backups"] == "number":
                self.radioButton_n_backups.setChecked(True)
            else:
                self.radioButton_smart_backups.setChecked(True)

        if "NOMAD_URL" in settings:
            self.lineEdit_oasis.setText(settings["NOMAD_URL"])
        else:
            self.lineEdit_oasis.setText(standard_pref["NOMAD_URL"])

        self.radioButton_mixed.clicked.connect(self.number_change)
        self.radioButton_plain_numbers.clicked.connect(self.number_change)
        self.radioButton_scientific.clicked.connect(self.number_change)
        if "extensions" in settings:
            self.extensions = settings["extensions"]
        else:
            self.extensions = standard_pref["extensions"]
        if "extension_path" in settings:
            self.pathButton_extension_path.set_path(settings["extension_path"])
        else:
            self.pathButton_extension_path.set_path(standard_pref["extension_path"])

        if "new_file_each_run" in settings:
            self.checkBox_new_file_each_run.setChecked(settings["new_file_each_run"])
        else:
            self.checkBox_new_file_each_run.setChecked(
                standard_pref["new_file_each_run"]
            )

        if "password_protection" in settings:
            self.checkBox_password.setChecked(settings["password_protection"])
        else:
            self.checkBox_password.setChecked(standard_pref["password_protection"])
        self.password_hash = ""
        if "password_hash" in settings:
            self.password_hash = settings["password_hash"]
        if "enable_API" in settings:
            self.checkBox_enable_Api.setChecked(settings["enable_API"])
            # Disable the generate and delete API keys push buttons if the API is disenabled
            if not self.checkBox_enable_Api.isChecked():
                self.pushButton_generate_Api_key.setEnabled(False)
                self.pushButton_copy_Api_key_clipboard.setEnabled(False)
                self.pushButton_delete_Api_keys.setEnabled(False)
        if "API_port" in settings:
            self.lineEdit_api_port.setText(settings["API_port"])
        self.pushButton_API_docu.clicked.connect(self.open_api_docu)
        self.pushButton_API_docu.setToolTip(
            "If the page does not load, accept your settings and try again. The API server might not be started yet."
        )

    def autosave_run_change(self):
        on = self.checkBox_autosave_run.isChecked()
        self.checkBox_backup_before_run.setEnabled(on)

    def number_change(self):
        mixed = self.radioButton_mixed.isChecked()
        self.spinBox_n_decimals.setEnabled(mixed)

    def accept(self):
        if self.password_hash:
            from nomad_camels.utility.password_widgets import Password_Dialog

            pw = Password_Dialog(parent=self, compare_hash=self.password_hash)
            if not pw.exec():
                return
        if self.checkBox_password.isChecked() and self.password_hash:
            msg_box = QMessageBox()
            msg_box.setWindowTitle("Change Password?")
            msg_box.setText("Do you want to change the password?")
            msg_box.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
            if msg_box.exec() != QMessageBox.Yes:
                super().accept()
        if self.checkBox_password.isChecked() and not self.password_hash:
            from nomad_camels.utility import password_widgets

            pw = password_widgets.Password_Dialog(parent=self, double_pass=True)
            if not pw.exec():
                return
            self.password_hash = password_widgets.hash_password(pw.password)

        super().accept()

    def change_theme(self):
        """ """
        theme = self.comboBox_theme.currentText()
        mat = theme == "qt-material"
        self.comboBox_material_theme.setEnabled(mat)
        material_theme = self.comboBox_material_theme.currentText()
        dark_mode = self.checkBox_dark.isChecked()
        change_theme(theme, material_theme=material_theme, dark_mode=dark_mode)

    def get_settings(self):
        """Reading all the UI-elements to get the selected settings,
        then returning those as a dictionary.

        Parameters
        ----------

        Returns
        -------

        """
        load_save_functions.update_config_path(self.pathButton_config_path.get_path())
        if not self.checkBox_password.isChecked():
            self.password_hash = None
        if self.radioButton_plain_numbers.isChecked():
            numbers = "plain"
        elif self.radioButton_scientific.isChecked():
            numbers = "scientific"
        else:
            numbers = "mixed"
        if self.radioButton_all_backups.isChecked():
            backups = "all"
        elif self.radioButton_n_backups.isChecked():
            backups = "number"
        else:
            backups = "smart"
        return {
            "autosave": self.checkBox_autosave.isChecked(),
            "autosave_run": self.checkBox_autosave_run.isChecked(),
            "backup_before_run": self.checkBox_backup_before_run.isChecked(),
            "dark_mode": self.checkBox_dark.isChecked(),
            "auto_check_updates": self.checkBox_auto_check_updates.isChecked(),
            "graphic_theme": self.comboBox_theme.currentText(),
            "material_theme": self.comboBox_material_theme.currentText(),
            "n_decimals": self.spinBox_n_decimals.value(),
            "number_format": numbers,
            "mixed_from": self.spinBox_scientific_from.value(),
            "py_files_path": self.pathButton_py_files.get_path(),
            "meas_files_path": self.pathButton_meas_files.get_path(),
            "device_driver_path": self.pathButton_device_path.get_path(),
            "databroker_catalog_name": self.lineEdit_catalog_name.text(),
            "number_databroker_files": self.spinBox_n_databroker_files.value(),
            "play_camel_on_error": self.checkBox_play_camel_on_error.isChecked(),
            "log_level": self.comboBox_log_level.currentText(),
            "logfile_size": self.spinBox_logfile_size.value(),
            "logfile_backups": self.spinBox_logfile_number.value(),
            "backups": backups,
            "NOMAD_URL": make_correct_url(self.lineEdit_oasis.text()),
            "password_protection": self.checkBox_password.isChecked(),
            "password_hash": self.password_hash,
            "new_file_each_run": self.checkBox_new_file_each_run.isChecked(),
            "extensions": self.extensions,
            "extension_path": self.pathButton_extension_path.get_path(),
            "enable_API": self.checkBox_enable_Api.isChecked(),
            "API_port": self.lineEdit_api_port.text(),
        }

    def keyPressEvent(self, a0: QKeyEvent) -> None:
        """Overwrites the keyPressEvent of the QDialog so that it does
        not close when pressing Enter/Return.

        Parameters
        ----------
        a0: QKeyEvent :


        Returns
        -------

        """
        if a0.key() == Qt.Key_Enter or a0.key() == Qt.Key_Return:
            return
        super().keyPressEvent(a0)

    def generate_api_key(self):
        # Generate a random API key (example: 40 characters long)
        api_key = secrets.token_urlsafe(40)
        self.Api_key_lineEdit.setText(api_key)
        # Save hash of the API key to the SQlite database file
        # Database setup
        data_base_path = os.path.join(load_save_functions.appdata_path, "CAMELS_API.db")
        conn = sqlite3.connect(data_base_path, check_same_thread=False)
        c = conn.cursor()
        c.execute(
            """
            CREATE TABLE IF NOT EXISTS api_keys (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                key TEXT NOT NULL,
                created_at DATETIME DEFAULT CURRENT_TIMESTAMP
            )
        """
        )
        conn.commit()
        # Store hashed API Key
        store_api_key(api_key, conn)
        conn.close()
        # Disable the "Generate API Key" button and enable the "Copy to Clipboard" button
        self.pushButton_generate_Api_key.setEnabled(False)
        self.pushButton_copy_Api_key_clipboard.setEnabled(True)
        # Turn the background of the copy_api_key_clipboard button to green
        self.pushButton_copy_Api_key_clipboard.setStyleSheet(
            "background-color: #0db002"
        )

    def confirm_delete_api_keys(self):
        reply = QMessageBox.question(
            self,
            "Confirm Action",
            "Are you sure you want to delete ALL API keys?\nYou can not reverse this action!\nIt will break all applications using API keys to communicate with CAMELS!",
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.No,
        )

        if reply == QMessageBox.Yes:
            data_base_path = os.path.join(
                load_save_functions.appdata_path, "CAMELS_API.db"
            )
            conn = sqlite3.connect(data_base_path, check_same_thread=False)
            c = conn.cursor()
            # SQL statement to drop the api_keys table
            c.execute("DROP TABLE IF EXISTS api_keys")
            conn.commit()
            conn.close()
        else:
            pass

    def copy_to_clipboard(self):
        api_key = self.Api_key_lineEdit.text()
        if api_key:
            clipboard = QApplication.clipboard()
            clipboard.setText(api_key)
            QMessageBox.information(self, "Copied", "API Key copied to clipboard!")
        else:
            QMessageBox.warning(self, "Warning", "No API Key to copy!")

    def change_enable_API(self):
        if self.checkBox_enable_Api.isChecked():
            self.pushButton_generate_Api_key.setEnabled(True)
            self.pushButton_delete_Api_keys.setEnabled(True)
            self.pushButton_API_docu.setEnabled(True)
        else:
            self.pushButton_generate_Api_key.setEnabled(False)
            self.pushButton_copy_Api_key_clipboard.setEnabled(False)
            self.pushButton_delete_Api_keys.setEnabled(False)
            self.pushButton_API_docu.setEnabled(False)
            self.pushButton_copy_Api_key_clipboard.setStyleSheet("")

    def open_api_docu(self):
        from nomad_camels.utility import variables_handling

        link = f"http://127.0.0.1:{self.lineEdit_api_port.text()}/docs"
        variables_handling.open_link(link)


def hash_api_key(api_key):
    return hashlib.sha256(api_key.encode()).hexdigest()


# Store API Key
def store_api_key(api_key, conn):
    hashed_key = hash_api_key(api_key)
    c = conn.cursor()
    c.execute("INSERT INTO api_keys (key) VALUES (?)", (hashed_key,))
    conn.commit()



from PySide6.QtCore import QItemSelectionModel, Qt


def test_install_demo_instrument(qtbot):
    """This test tries to install the demo device, using the
    instrument_installer widget."""
    from nomad_camels.frontpanels import instrument_installer

    installer = instrument_installer.Instrument_Installer()
    qtbot.addWidget(
        installer
    )  # used to run the qt-app and close the widget after the test
    install_demo(qtbot, installer)


def install_demo(qtbot, installer):
    """A helper function, used by `test_install_demo_instrument` and also by
    `test_add_device`, if the device is not yet installed"""
    for box in installer.checkboxes:
        if box.text() == "demo_instrument":
            box.setChecked(True)
            break
    with qtbot.waitSignal(installer.instruments_updated, timeout=60000) as blocker:
        qtbot.mouseClick(
            installer.pushButton_install_update_selected, Qt.MouseButton.LeftButton
        )
        # the instruments_updated signal is being waited for, after clicking on
        # the install/update button
    assert "demo_instrument" in installer.installed_devs


def test_add_device(qtbot):
    """Going through the instrument management, if demo_instrument is not installed,
    it will be. Then the config widget is used to add a demo_instrument.
    In the end there is a check whether it is in the list of devices."""
    from nomad_camels.frontpanels import manage_instruments

    manager = manage_instruments.ManageInstruments()
    qtbot.addWidget(manager)
    installer = manager.installer
    if "demo_instrument" not in installer.installed_devs:
        install_demo(qtbot, installer)  # install if not there
    conf = manager.config_widget
    conf.build_table()
    item1 = None
    item2 = None
    for row in range(conf.tableWidget_instruments.rowCount()):
        item1 = conf.tableWidget_instruments.item(row, 0)
        item2 = conf.tableWidget_instruments.item(row, 1)
        if item1.text() == "demo_instrument":
            break
    assert item1 is not None  # demo_instrument should be in the installed table
    assert item2 is not None
    index1 = conf.tableWidget_instruments.indexFromItem(item1)
    conf.tableWidget_instruments.selectionModel().select(
        index1, QItemSelectionModel.Select
    )
    index2 = conf.tableWidget_instruments.indexFromItem(item2)
    conf.tableWidget_instruments.selectionModel().select(
        index2, QItemSelectionModel.Select
    )
    # two items need to be selected, since with a real click, the selection also
    # returns a list
    conf.table_click()  # select

    def check_instr_in():
        """ """
        # clicking on an added device, check whether it is actually added
        qtbot.mouseClick(conf.pushButton_add, Qt.MouseButton.LeftButton)
        instr = conf.get_config()
        assert "demo_instrument" in instr

    qtbot.waitUntil(check_instr_in)  # wait for the qt event loops



# Resource object code (Python 3)
# Created by: object code
# Created by: The Resource Compiler for Qt version 6.6.1
# WARNING! All changes made in this file will be lost!

from PySide6 import QtCore

qt_resource_data = b"\
\x00\x00\x08\xc0\
<\
?xml version=\x221.\
0\x22 encoding=\x22UTF\
-8\x22 standalone=\x22\
no\x22?>\x0a<!-- Creat\
ed with Inkscape\
 (http://www.ink\
scape.org/) -->\x0a\
\x0a<svg\x0a   width=\x22\
24.853426mm\x22\x0a   \
height=\x2229.09959\
8mm\x22\x0a   viewBox=\
\x220 0 24.853426 2\
9.099598\x22\x0a   ver\
sion=\x221.1\x22\x0a   id\
=\x22svg1\x22\x0a   inksc\
ape:version=\x221.3\
 (0e150ed6c4, 20\
23-07-21)\x22\x0a   so\
dipodi:docname=\x22\
copy_to_clipboar\
d.svg\x22\x0a   xmlns:\
inkscape=\x22http:/\
/www.inkscape.or\
g/namespaces/ink\
scape\x22\x0a   xmlns:\
sodipodi=\x22http:/\
/sodipodi.source\
forge.net/DTD/so\
dipodi-0.dtd\x22\x0a  \
 xmlns=\x22http://w\
ww.w3.org/2000/s\
vg\x22\x0a   xmlns:svg\
=\x22http://www.w3.\
org/2000/svg\x22>\x0a \
 <sodipodi:named\
view\x0a     id=\x22na\
medview1\x22\x0a     p\
agecolor=\x22#fffff\
f\x22\x0a     borderco\
lor=\x22#000000\x22\x0a  \
   borderopacity\
=\x220.25\x22\x0a     ink\
scape:showpagesh\
adow=\x222\x22\x0a     in\
kscape:pageopaci\
ty=\x220.0\x22\x0a     in\
kscape:pagecheck\
erboard=\x220\x22\x0a    \
 inkscape:deskco\
lor=\x22#d1d1d1\x22\x0a  \
   inkscape:docu\
ment-units=\x22mm\x22\x0a\
     inkscape:zo\
om=\x221.6017536\x22\x0a \
    inkscape:cx=\
\x22-29.342839\x22\x0a   \
  inkscape:cy=\x221\
13.31331\x22\x0a     i\
nkscape:window-w\
idth=\x221920\x22\x0a    \
 inkscape:window\
-height=\x221137\x22\x0a \
    inkscape:win\
dow-x=\x221912\x22\x0a   \
  inkscape:windo\
w-y=\x22-8\x22\x0a     in\
kscape:window-ma\
ximized=\x221\x22\x0a    \
 inkscape:curren\
t-layer=\x22layer1\x22\
 />\x0a  <defs\x0a    \
 id=\x22defs1\x22 />\x0a \
 <g\x0a     inkscap\
e:label=\x22Layer 1\
\x22\x0a     inkscape:\
groupmode=\x22layer\
\x22\x0a     id=\x22layer\
1\x22\x0a     transfor\
m=\x22translate(-71\
.731018,-117.893\
09)\x22>\x0a    <rect\x0a\
       style=\x22op\
acity:1;fill:non\
e;stroke:#000000\
;stroke-width:2;\
stroke-dasharray\
:none\x22\x0a       id\
=\x22rect1\x22\x0a       \
width=\x2216.568283\
\x22\x0a       height=\
\x2220.826275\x22\x0a    \
   x=\x2272.731018\x22\
\x0a       y=\x22125.1\
6641\x22 />\x0a    <pa\
th\x0a       style=\
\x22opacity:1;fill:\
#000000;stroke:#\
000000;stroke-wi\
dth:2;stroke-das\
harray:none\x22\x0a   \
    d=\x22m 95.5844\
44,123.79879 v 1\
6.11873 l -15.65\
1524,-0.1168 v -\
20.90763 h 10.16\
1811 z\x22\x0a       i\
d=\x22path1\x22\x0a      \
 sodipodi:nodety\
pes=\x22cccccc\x22 />\x0a\
    <path\x0a      \
 style=\x22opacity:\
1;fill:#000000;s\
troke:#000000;st\
roke-width:2;str\
oke-dasharray:no\
ne\x22\x0a       d=\x22M \
75.695423,131.76\
851 H 81.14648\x22\x0a\
       id=\x22path2\
\x22 />\x0a    <path\x0a \
      style=\x22fil\
l:#000000;stroke\
:#000000;stroke-\
width:2;stroke-d\
asharray:none\x22\x0a \
      d=\x22m 75.77\
4241,137.26849 h\
 5.451057\x22\x0a     \
  id=\x22path2-0\x22 /\
>\x0a    <path\x0a    \
   style=\x22fill:#\
000000;stroke:#0\
00000;stroke-wid\
th:2;stroke-dash\
array:none\x22\x0a    \
   d=\x22m 75.86494\
7,142.76847 h 9.\
085095\x22\x0a       i\
d=\x22path2-0-6\x22 />\
\x0a  </g>\x0a</svg>\x0a\
"

qt_resource_name = b"\
\x00\x06\
\x07\x03}\xc3\
\x00i\
\x00m\x00a\x00g\x00e\x00s\
\x00\x08\
\x08\x86\xe2c\
\x00g\
\x00r\x00a\x00p\x00h\x00i\x00c\x00s\
\x00\x15\
\x0b>\x0e\x07\
\x00c\
\x00o\x00p\x00y\x00_\x00t\x00o\x00_\x00c\x00l\x00i\x00p\x00b\x00o\x00a\x00r\x00d\
\x00.\x00s\x00v\x00g\
"

qt_resource_struct = b"\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x01\
\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x02\
\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x00\x00\x12\x00\x02\x00\x00\x00\x01\x00\x00\x00\x03\
\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x00\x00(\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\
\x00\x00\x01\x90\xc0\x1c\xf1\xd7\
"

def qInitResources():
    QtCore.qRegisterResourceData(0x03, qt_resource_struct, qt_resource_name, qt_resource_data)

def qCleanupResources():
    QtCore.qUnregisterResourceData(0x03, qt_resource_struct, qt_resource_name, qt_resource_data)

qInitResources()



import numpy as np

from bluesky.callbacks.core import CallbackBase

from PySide6.QtWidgets import QTableWidgetItem, QTableWidget, QWidget, QGridLayout
from PySide6.QtCore import Qt, QEvent, QObject
from PySide6.QtGui import QIcon

from nomad_camels.bluesky_handling.evaluation_helper import Evaluator

from nomad_camels.utility.plot_placement import place_widget
from importlib import resources
from nomad_camels import graphics

from PySide6.QtCore import Signal as pySignal


class Values_List_Plot(QWidget):
    """ """

    closing = pySignal()
    reopened = pySignal()

    def __init__(
        self,
        value_list,
        *,
        epoch="run",
        namespace=None,
        title="",
        stream_name="primary",
        parent=None,
        plot_all_available=True,
        **kwargs,
    ):
        super().__init__(parent=parent)
        self.table = QTableWidget()
        self.livePlot = Live_List(
            value_list,
            epoch=epoch,
            namespace=namespace,
            stream_name=stream_name,
            parent=self,
            table=self.table,
            plot_all_available=plot_all_available,
            **kwargs,
        )
        self.livePlot.new_data.connect(self.show_again)

        layout = QGridLayout()
        layout.addWidget(self.table)
        self.setLayout(layout)
        if title:
            self.setWindowTitle(title)
        elif value_list:
            self.setWindowTitle(f"{value_list[0]} ...")
        else:
            self.setWindowTitle("Current Value List")
        self.setWindowIcon(QIcon(str(resources.files(graphics) / "camels_icon.png")))
        self.stream_name = stream_name
        place_widget(self)

    def show_again(self):
        if not self.isVisible():
            self.show()
            self.reopened.emit()

    def clear_plot(self):
        pass

    def autoscale(self):
        pass

    def closeEvent(self, a0):
        """

        Parameters
        ----------
        a0 :


        Returns
        -------

        """
        self.closing.emit()
        super().closeEvent(a0)


class Teleporter(QObject):
    name_doc_escape = pySignal(str, dict, object)


def handle_teleport(name, doc, obj):
    obj(name, doc, escape=True)


class Live_List(QObject, CallbackBase):
    new_data = pySignal()

    def __init__(
        self,
        value_list,
        table,
        *,
        epoch="run",
        namespace=None,
        stream_name="primary",
        parent=None,
        plot_all_available=False,
        **kwargs,
    ):
        CallbackBase.__init__(self)
        QObject.__init__(self)
        self.__teleporter = Teleporter()
        self.__teleporter.name_doc_escape.connect(handle_teleport)
        if isinstance(value_list, str):
            value_list = [value_list]
        self.value_list = value_list
        self.stream_name = stream_name
        self.table = table
        self.plot_all_available = plot_all_available
        self.eva = Evaluator(namespace=namespace)

        self.table.setRowCount(len(value_list))
        self.table.setColumnCount(2)
        self.table.verticalHeader().setHidden(True)
        self.table.horizontalHeader().setHidden(True)
        self.val_items = []
        for i, val in enumerate(value_list):
            item = QTableWidgetItem(val)
            item.setFlags(item.flags() & ~Qt.ItemIsEditable)
            self.table.setItem(i, 0, item)
            item = QTableWidgetItem(str(np.nan))
            item.setFlags(item.flags() & ~Qt.ItemIsEditable)
            self.table.setItem(i, 1, item)
            self.val_items.append(item)
        self.add_items = {}
        self.desc = ""
        self.epoch_offset = 0
        self.epoch = epoch
        self.table.resizeColumnsToContents()

    def add_to_table(self, name):
        i = len(self.val_items) + len(self.add_items)
        self.table.setRowCount(i + 1)
        item = QTableWidgetItem(name)
        item.setFlags(item.flags() & ~Qt.ItemIsEditable)
        self.table.setItem(i, 0, item)
        item = QTableWidgetItem(str(np.nan))
        item.setFlags(item.flags() & ~Qt.ItemIsEditable)
        self.table.setItem(i, 1, item)
        self.add_items[name] = item

    def descriptor(self, doc):
        """

        Parameters
        ----------
        doc :


        Returns
        -------

        """
        if doc["name"] == self.stream_name:
            self.desc = doc["uid"]

    def start(self, doc):
        """

        Parameters
        ----------
        doc :


        Returns
        -------

        """
        self.epoch_offset = doc["time"]
        super().start(doc)

    def event(self, doc):
        """

        Parameters
        ----------
        doc :


        Returns
        -------

        """
        if isinstance(doc, QEvent):
            return QWidget.event(self, doc)
        if doc["descriptor"] != self.desc:
            return

        for i, val in enumerate(self.value_list):
            if val == "time":
                new_val = doc["time"]
                if self.epoch == "run":
                    new_val -= self.epoch_offset
            else:
                try:
                    new_val = doc["data"][val]
                except KeyError:
                    if not self.eva.is_to_date(doc["time"]):
                        self.eva.event(doc)
                    new_val = self.eva.eval(val)
            if isinstance(new_val, bool) or np.issubdtype(type(new_val), np.bool_):
                self.val_items[i].setText(str(new_val))
            else:
                self.val_items[i].setText(f"{new_val:7e}")
        if self.plot_all_available:
            for name, value in doc["data"].items():
                if name in self.value_list:
                    continue
                if name not in self.add_items:
                    self.add_to_table(name)
                try:
                    self.add_items[name].setText(f"{value:7e}")
                except:
                    self.add_items[name].setText(str(value))
        self.table.resizeColumnsToContents()
        self.new_data.emit()

    def __call__(self, name, doc, *, escape=False):
        """
        The call method of the callback. This method is called when a new event is received. If `__teleporter` is set, the event is sent to the `__teleporter`, otherwise the event is processed directly. The teleporter is necessary to send the event to a different thread, since Qt objects can only be accessed from the thread they were created in.

        Parameters
        ----------
        name : str
            The name of the event
        doc : dict
            The event document
        escape : bool, (default: False)
            If True, the event is always processed directly, otherwise it is sent to the teleporter
        """
        if not escape and self.__teleporter is not None:
            self.__teleporter.name_doc_escape.emit(name, doc, self)
        else:
            return CallbackBase.__call__(self, name, doc)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'enterTextDialog.ui'
##
## Created by: Qt User Interface Compiler version 6.5.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QAbstractButton,
    QApplication,
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QLabel,
    QLineEdit,
    QSizePolicy,
    QWidget,
)


class Ui_EnterTextDialog(object):
    """ """

    def setupUi(self, EnterTextDialog):
        """

        Parameters
        ----------
        EnterTextDialog :


        Returns
        -------

        """
        if not EnterTextDialog.objectName():
            EnterTextDialog.setObjectName("EnterTextDialog")
        EnterTextDialog.resize(278, 106)
        self.gridLayout_2 = QGridLayout(EnterTextDialog)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.gridLayout = QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.lineEdit_text = QLineEdit(EnterTextDialog)
        self.lineEdit_text.setObjectName("lineEdit_text")

        self.gridLayout.addWidget(self.lineEdit_text, 0, 1, 1, 1)

        self.label = QLabel(EnterTextDialog)
        self.label.setObjectName("label")

        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)

        self.gridLayout_2.addLayout(self.gridLayout, 0, 0, 1, 1)

        self.buttonBox = QDialogButtonBox(EnterTextDialog)
        self.buttonBox.setObjectName("buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)

        self.gridLayout_2.addWidget(self.buttonBox, 1, 0, 1, 1)

        self.retranslateUi(EnterTextDialog)
        self.buttonBox.accepted.connect(EnterTextDialog.accept)
        self.buttonBox.rejected.connect(EnterTextDialog.reject)

        QMetaObject.connectSlotsByName(EnterTextDialog)

    # setupUi

    def retranslateUi(self, EnterTextDialog):
        """

        Parameters
        ----------
        EnterTextDialog :


        Returns
        -------

        """
        EnterTextDialog.setWindowTitle(
            QCoreApplication.translate("EnterTextDialog", "Dialog", None)
        )
        self.label.setText(
            QCoreApplication.translate("EnterTextDialog", "Task name:", None)
        )

    # retranslateUi



from PySide6.QtWidgets import QDialog

from nomad_camels.gui.pass_ask import Ui_Pass_Ask


class Pass_Ask(Ui_Pass_Ask, QDialog):
    """ """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setupUi(self)



import sys
import os
import ast

from PySide6.QtWidgets import (
    QDialog,
    QLabel,
    QPushButton,
    QFileDialog,
    QMessageBox,
    QLineEdit,
    QGridLayout,
)
from PySide6.QtGui import QIcon, QCloseEvent

from nomad_camels.ui_widgets.add_remove_table import AddRemoveTable
from nomad_camels.utility import variables_handling, fit_variable_renaming

from importlib import resources
from nomad_camels import graphics


class EPICS_Driver_Builder(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)

        try:
            self.setWindowIcon(
                QIcon(str(resources.files(graphics) / "camels_icon.png"))
            )
        except:
            pass
        self.setWindowTitle("EPICS-driver-builder - NOMAD CAMELS")

        label_name = QLabel("Instrument Name:")
        self.lineEdit_name = QLineEdit()

        label_in = QLabel("Input Channels")
        label_out = QLabel("Output Channels")
        label_config = QLabel("Config Channels")
        label_config_in = QLabel("Config Channels - Read Only")
        label_in.setStyleSheet("font-weight: bold")
        label_out.setStyleSheet("font-weight: bold")
        label_config.setStyleSheet("font-weight: bold")
        label_config_in.setStyleSheet("font-weight: bold")

        channel_info_config = ["PV-Name", "Datatype"]
        channel_info = ["PV-Name"]
        comboBoxes = {"Datatype": ["str", "float", "bool"]}
        self.table_in = AddRemoveTable(headerLabels=channel_info)
        self.table_out = AddRemoveTable(headerLabels=channel_info)
        self.table_config = AddRemoveTable(
            headerLabels=channel_info_config, comboBoxes=comboBoxes
        )
        self.table_config_in = AddRemoveTable(headerLabels=channel_info)

        self.button_build = QPushButton("Build Driver")
        self.button_cancel = QPushButton("Cancel")
        self.button_cancel.clicked.connect(self.close)
        self.button_build.clicked.connect(self.build_device)

        layout = QGridLayout()
        layout.addWidget(label_name, 0, 0)
        layout.addWidget(self.lineEdit_name, 0, 1, 1, 3)
        layout.addWidget(label_in, 10, 0, 1, 2)
        layout.addWidget(self.table_in, 11, 0, 1, 2)
        layout.addWidget(label_out, 10, 2, 1, 2)
        layout.addWidget(self.table_out, 11, 2, 1, 2)
        layout.addWidget(label_config_in, 30, 0, 1, 2)
        layout.addWidget(self.table_config_in, 31, 0, 1, 2)
        layout.addWidget(label_config, 30, 2, 1, 2)
        layout.addWidget(self.table_config, 31, 2, 1, 2)
        layout.addWidget(self.button_build, 50, 2)
        layout.addWidget(self.button_cancel, 50, 3)
        self.setLayout(layout)
        # self.resize(600, 700)

    def build_device(self):
        dev_name = fit_variable_renaming.replace_name(self.lineEdit_name.text())
        inputs = self.table_in.update_table_data()
        outputs = self.table_out.update_table_data()
        configs = self.table_config.update_table_data()
        configs_in = self.table_config_in.update_table_data()

        path = (
            variables_handling.device_driver_path
            or QFileDialog.getExistingDirectory(self, "Select directory for drivers")
        )
        if not path:
            return

        def check_builtin(pv):
            if pv in vars(__builtins__):
                raise Exception(
                    f'PV-name "{pv}" resembles python builtin function or variable!\nDefine another name!\nYou may change the PV-name later in the code, see the documentation for more information.'
                )
            try:
                ast.parse(f"{pv} = None")
            except (ValueError, SyntaxError, TypeError):
                raise Exception(
                    f'PV-name "{pv}" resembles python builtin function or variable!\nDefine another name!\nYou may change the PV-name later in the code, see the documentation for more information.'
                )

        all_good = True
        for name in inputs["PV-Name"]:
            all_good = all_good and variables_handling.check_variable_name(
                name, parent=self
            )
        for name in outputs["PV-Name"]:
            all_good = all_good and variables_handling.check_variable_name(
                name, parent=self
            )
        for name in configs["PV-Name"]:
            all_good = all_good and variables_handling.check_variable_name(
                name, parent=self
            )
        for name in configs_in["PV-Name"]:
            all_good = all_good and variables_handling.check_variable_name(
                name, parent=self
            )
        if not all_good:
            raise Exception(
                f"Some PV-name resemble python builtin function or variable!\nDefine another name!\nYou may change the PV-name later in the code, see the documentation for more information."
            )
        directory = f"{path}/nomad_camels_driver_{dev_name}"
        sys.path.append(directory)
        fname = f"{directory}/{dev_name}.py"
        if os.path.isdir(directory) and os.path.isfile(fname):
            answer = QMessageBox.question(
                self,
                "Instrument already exists!",
                f"Instrument type with name {dev_name} already exists.\nDo you want to overwrite the existing instrument driver?",
                buttons=QMessageBox.Yes | QMessageBox.Cancel,
                defaultButton=QMessageBox.Cancel,
            )
            if answer != QMessageBox.Yes:
                return
        if not os.path.isdir(directory):
            os.mkdir(directory)

        class_string = f"from nomad_camels_driver_{dev_name}.{dev_name}_ophyd import {dev_name}_Ophyd\n\n"
        class_string += "from nomad_camels.main_classes import device_class\n\n"
        class_string += "class subclass(device_class.Device):\n"
        class_string += "\tdef __init__(self, **kwargs):\n"
        class_string += f'\t\tsuper().__init__(name="{dev_name}", virtual=False, directory="{dev_name}", ophyd_device={dev_name}_Ophyd, ophyd_class_name="{dev_name}_Ophyd", **kwargs)\n'
        for i, name in enumerate(configs["PV-Name"]):
            if configs["Datatype"][i] == "str":
                class_string += f'\t\tself.config["{name}"] = ""\n'
            elif configs["Datatype"][i] == "float":
                class_string += f'\t\tself.config["{name}"] = 0\n'
            elif configs["Datatype"][i] == "bool":
                class_string += f'\t\tself.config["{name}"] = False\n'
        class_string += "\n\nclass subclass_config(device_class.Simple_Config):\n"
        class_string += '\tdef __init__(self, parent=None, data="", settings_dict=None, config_dict=None, additional_info=None):\n'
        class_string += f'\t\tsuper().__init__(parent, "{dev_name}", data, settings_dict, config_dict, additional_info)\n'
        class_string += '\t\tself.comboBox_connection_type.addItem("EPICS")\n'
        class_string += "\t\tself.load_settings()\n"

        ophyd_string = "from ophyd import Component as Cpt\n\n"
        ophyd_string += "from ophyd import Device, EpicsSignal, EpicsSignalRO\n\n"
        ophyd_string += f"class {dev_name}_Ophyd(Device):\n"
        for i, name in enumerate(inputs["PV-Name"]):
            ophyd_string += f'\t{name} = Cpt(EpicsSignalRO, "{name}")\n'
        for i, name in enumerate(outputs["PV-Name"]):
            ophyd_string += f'\t{name} = Cpt(EpicsSignal, "{name}")\n'
        for i, name in enumerate(configs["PV-Name"]):
            ophyd_string += f'\t{name} = Cpt(EpicsSignal, "{name}",kind="config")\n'
        for i, name in enumerate(configs_in["PV-Name"]):
            ophyd_string += f'\t{name} = Cpt(EpicsSignalRO, "{name}", kind="config")\n'

        ophyd_string += f'\n\tdef __init__(self, prefix="", *, name, kind=None, read_attrs=None, configuration_attrs=None, parent=None, **kwargs):\n'
        ophyd_string += f"\t\tsuper().__init__(prefix=prefix, name=name, kind=kind, read_attrs=read_attrs, configuration_attrs=configuration_attrs, parent=parent, **kwargs)"

        with open(f"{directory}/{dev_name}.py", "w", encoding="utf-8") as f:
            f.write(class_string)
        with open(f"{directory}/{dev_name}_ophyd.py", "w", encoding="utf-8") as f:
            f.write(ophyd_string)
        # with open(f'{directory}/__init__.py', 'w') as f:
        #     f.write(ophyd_string)

        QMessageBox.information(
            self, "Build finished!", f"The instrument-driver {dev_name} has been built!"
        )
        self.accept()

    def closeEvent(self, a0: QCloseEvent) -> None:
        discard_dialog = QMessageBox.question(
            self,
            "Discard Changes?",
            f"All changes will be lost!",
            QMessageBox.Yes | QMessageBox.No,
        )
        if discard_dialog != QMessageBox.Yes:
            a0.ignore()
            return
        super().closeEvent(a0)


if __name__ == "__main__":
    from PySide6.QtWidgets import QApplication
    from PySide6.QtCore import QCoreApplication
    from nomad_camels.utility import exception_hook

    sys.excepthook = exception_hook.exception_hook
    app = QCoreApplication.instance()
    if app is None:
        app = QApplication(sys.argv)
    ui = EPICS_Driver_Builder()
    ui.show()
    app.exec()
